import base64
import hashlib
import hmac
import json
import logging
import math
import secrets
import string
import time
from collections.abc import Callable
from dataclasses import dataclass

import aiohttp
from aiohttp import ContentTypeError, FormData
from pyrate_limiter import Duration, Limiter, Rate

from roborock import HomeDataSchedule
from roborock.data import HomeData, HomeDataRoom, HomeDataScene, ProductResponse, RRiot, UserData
from roborock.exceptions import (
    RoborockAccountDoesNotExist,
    RoborockException,
    RoborockInvalidCode,
    RoborockInvalidCredentials,
    RoborockInvalidEmail,
    RoborockInvalidUserAgreement,
    RoborockMissingParameters,
    RoborockNoResponseFromBaseURL,
    RoborockNoUserAgreement,
    RoborockRateLimit,
    RoborockTooFrequentCodeRequests,
)

_LOGGER = logging.getLogger(__name__)
BASE_URLS = [
    "https://usiot.roborock.com",
    "https://euiot.roborock.com",
    "https://cniot.roborock.com",
    "https://ruiot.roborock.com",
]


@dataclass
class IotLoginInfo:
    """Information about the login to the iot server."""

    base_url: str
    country_code: str
    country: str


class RoborockApiClient:
    _LOGIN_RATES = [
        Rate(1, Duration.SECOND),
        Rate(3, Duration.MINUTE),
        Rate(10, Duration.HOUR),
        Rate(20, Duration.DAY),
    ]
    _HOME_DATA_RATES = [
        Rate(1, Duration.SECOND),
        Rate(3, Duration.MINUTE),
        Rate(5, Duration.HOUR),
        Rate(40, Duration.DAY),
    ]

    _login_limiter = Limiter(_LOGIN_RATES)
    _home_data_limiter = Limiter(_HOME_DATA_RATES)

    def __init__(
        self, username: str, base_url: str | None = None, session: aiohttp.ClientSession | None = None
    ) -> None:
        """Sample API Client."""
        self._username = username
        self._base_url = base_url
        self._device_identifier = secrets.token_urlsafe(16)
        self.session = session
        self._iot_login_info: IotLoginInfo | None = None
        self._base_urls = BASE_URLS if base_url is None else [base_url]

    async def _get_iot_login_info(self) -> IotLoginInfo:
        if self._iot_login_info is None:
            for iot_url in self._base_urls:
                url_request = PreparedRequest(iot_url, self.session)
                response = await url_request.request(
                    "post",
                    "/api/v1/getUrlByEmail",
                    params={"email": self._username, "needtwostepauth": "false"},
                )
                if response is None:
                    continue
                response_code = response.get("code")
                if response_code != 200:
                    if response_code == 2003:
                        raise RoborockInvalidEmail("Your email was incorrectly formatted.")
                    elif response_code == 1001:
                        raise RoborockMissingParameters(
                            "You are missing parameters for this request, are you sure you entered your username?"
                        )
                    else:
                        raise RoborockException(f"{response.get('msg')} - response code: {response_code}")
                country_code = response["data"]["countrycode"]
                country = response["data"]["country"]
                if country_code is not None or country is not None:
                    self._iot_login_info = IotLoginInfo(
                        base_url=response["data"]["url"],
                        country=country,
                        country_code=country_code,
                    )
                    _LOGGER.debug("Country determined to be %s and code is %s", country, country_code)
                    return self._iot_login_info
            raise RoborockNoResponseFromBaseURL(
                "No account was found for any base url we tried. Either your email is incorrect or we do not have a"
                " record of the roborock server your device is on."
            )
        return self._iot_login_info

    @property
    async def base_url(self):
        if self._base_url is not None:
            return self._base_url
        return (await self._get_iot_login_info()).base_url

    @property
    async def country(self):
        return (await self._get_iot_login_info()).country

    @property
    async def country_code(self):
        return (await self._get_iot_login_info()).country_code

    def _get_header_client_id(self):
        md5 = hashlib.md5()
        md5.update(self._username.encode())
        md5.update(self._device_identifier.encode())
        return base64.b64encode(md5.digest()).decode()

    async def nc_prepare(self, user_data: UserData, timezone: str) -> dict:
        """This gets a few critical parameters for adding a device to your account."""
        if (
            user_data.rriot is None
            or user_data.rriot.r is None
            or user_data.rriot.u is None
            or user_data.rriot.r.a is None
        ):
            raise RoborockException("Your userdata is missing critical attributes.")
        base_url = user_data.rriot.r.a
        prepare_request = PreparedRequest(base_url, self.session)
        hid = await self._get_home_id(user_data)

        data = FormData()
        data.add_field("hid", hid)
        data.add_field("tzid", timezone)

        prepare_response = await prepare_request.request(
            "post",
            "/nc/prepare",
            headers={
                "Authorization": _get_hawk_authentication(
                    user_data.rriot, "/nc/prepare", {"hid": hid, "tzid": timezone}
                ),
            },
            data=data,
        )

        if prepare_response is None:
            raise RoborockException("prepare_response is None")
        if not prepare_response.get("success"):
            raise RoborockException(f"{prepare_response.get('msg')} - response code: {prepare_response.get('code')}")

        return prepare_response["result"]

    async def add_device(self, user_data: UserData, s: str, t: str) -> dict:
        """This will add a new device to your account
        it is recommended to only use this during a pairing cycle with a device.
        Please see here: https://github.com/Python-roborock/Roborockmitmproxy/blob/main/handshake_protocol.md
        """
        if (
            user_data.rriot is None
            or user_data.rriot.r is None
            or user_data.rriot.u is None
            or user_data.rriot.r.a is None
        ):
            raise RoborockException("Your userdata is missing critical attributes.")
        base_url = user_data.rriot.r.a
        add_device_request = PreparedRequest(base_url, self.session)

        add_device_response = await add_device_request.request(
            "GET",
            "/user/devices/newadd",
            headers={
                "Authorization": _get_hawk_authentication(
                    user_data.rriot, "/user/devices/newadd", params={"s": s, "t": t}
                ),
            },
            params={"s": s, "t": t},
        )

        if add_device_response is None:
            raise RoborockException("add_device is None")
        if not add_device_response.get("success"):
            raise RoborockException(
                f"{add_device_response.get('msg')} - response code: {add_device_response.get('code')}"
            )

        return add_device_response["result"]

    async def request_code(self) -> None:
        if not await self._login_limiter.try_acquire_async("login", blocking=True, timeout=1):
            raise RoborockRateLimit("Reached maximum requests for login. Please try again later.")
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()
        code_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})

        code_response = await code_request.request(
            "post",
            "/api/v1/sendEmailCode",
            params={
                "username": self._username,
                "type": "auth",
            },
        )
        if code_response is None:
            raise RoborockException("Failed to get a response from send email code")
        response_code = code_response.get("code")
        if response_code != 200:
            _LOGGER.info("Request code failed for %s with the following context: %s", self._username, code_response)
            if response_code == 2008:
                raise RoborockAccountDoesNotExist("Account does not exist - check your login and try again.")
            elif response_code == 9002:
                raise RoborockTooFrequentCodeRequests("You have attempted to request too many codes. Try again later")
            else:
                raise RoborockException(f"{code_response.get('msg')} - response code: {code_response.get('code')}")

    async def request_code_v4(self) -> None:
        """Request a code using the v4 endpoint."""
        if await self.country_code is None or await self.country is None:
            _LOGGER.info("No country code or country found, trying old version of request code.")
            return await self.request_code()
        if not await self._login_limiter.try_acquire_async("login", blocking=True, timeout=1):
            raise RoborockRateLimit("Reached maximum requests for login. Please try again later.")
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()
        code_request = PreparedRequest(
            base_url,
            self.session,
            {
                "header_clientid": header_clientid,
                "Content-Type": "application/x-www-form-urlencoded",
                "header_clientlang": "en",
            },
        )

        code_response = await code_request.request(
            "post",
            "/api/v4/email/code/send",
            data={"email": self._username, "type": "login", "platform": ""},
        )
        if code_response is None:
            raise RoborockException("Failed to get a response from send email code")
        response_code = code_response.get("code")
        if response_code != 200:
            _LOGGER.info("Request code failed for %s with the following context: %s", self._username, code_response)
            if response_code == 2008:
                raise RoborockAccountDoesNotExist("Account does not exist - check your login and try again.")
            elif response_code == 9002:
                raise RoborockTooFrequentCodeRequests("You have attempted to request too many codes. Try again later")
            elif response_code == 3030 and len(self._base_urls) > 1:
                self._base_urls = self._base_urls[1:]
                self._iot_login_info = None
                return await self.request_code_v4()
            else:
                raise RoborockException(f"{code_response.get('msg')} - response code: {code_response.get('code')}")

    async def _sign_key_v3(self, s: str) -> str:
        """Sign a randomly generated string."""
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()
        code_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})

        code_response = await code_request.request(
            "post",
            "/api/v3/key/sign",
            params={"s": s},
        )

        if not code_response or "data" not in code_response or "k" not in code_response["data"]:
            raise RoborockException("Failed to get a response from sign key")
        response_code = code_response.get("code")

        if response_code != 200:
            _LOGGER.info("Request code failed for %s with the following context: %s", self._username, code_response)
            raise RoborockException(f"{code_response.get('msg')} - response code: {code_response.get('code')}")

        return code_response["data"]["k"]

    async def code_login_v4(
        self, code: int | str, country: str | None = None, country_code: int | None = None
    ) -> UserData:
        """
        Login via code authentication.
        :param code: The code from the email.
        :param country: The two-character representation of the country, i.e. "US"
        :param country_code: the country phone number code i.e. 1 for US.
        """
        base_url = await self.base_url
        if country is None:
            country = await self.country
        if country_code is None:
            country_code = await self.country_code
        if country_code is None or country is None:
            _LOGGER.info("No country code or country found, trying old version of code login.")
            return await self.code_login(code)
        header_clientid = self._get_header_client_id()
        x_mercy_ks = "".join(secrets.choice(string.ascii_letters + string.digits) for _ in range(16))
        x_mercy_k = await self._sign_key_v3(x_mercy_ks)
        login_request = PreparedRequest(
            base_url,
            self.session,
            {
                "header_clientid": header_clientid,
                "x-mercy-ks": x_mercy_ks,
                "x-mercy-k": x_mercy_k,
                "Content-Type": "application/x-www-form-urlencoded",
                "header_clientlang": "en",
                "header_appversion": "4.54.02",
                "header_phonesystem": "iOS",
                "header_phonemodel": "iPhone16,1",
            },
        )
        login_response = await login_request.request(
            "post",
            "/api/v4/auth/email/login/code",
            data={
                "country": country,
                "countryCode": country_code,
                "email": self._username,
                "code": code,
                # Major and minor version are the user agreement version, we will need to see if this needs to be
                # dynamic https://usiot.roborock.com/api/v3/app/agreement/latest?country=US
                "majorVersion": 14,
                "minorVersion": 0,
            },
        )
        if login_response is None:
            raise RoborockException("Login request response is None")
        response_code = login_response.get("code")
        if response_code != 200:
            _LOGGER.info("Login failed for %s with the following context: %s", self._username, login_response)
            if response_code == 2018:
                raise RoborockInvalidCode("Invalid code - check your code and try again.")
            if response_code == 3009:
                raise RoborockNoUserAgreement("You must accept the user agreement in the Roborock app to continue.")
            if response_code == 3006:
                raise RoborockInvalidUserAgreement(
                    "User agreement must be accepted again - or you are attempting to use the Mi Home app account."
                )
            if response_code == 3039:
                raise RoborockAccountDoesNotExist(
                    "This account does not exist - please ensure that you selected the right region and email."
                )
            raise RoborockException(f"{login_response.get('msg')} - response code: {response_code}")
        user_data = login_response.get("data")
        if not isinstance(user_data, dict):
            raise RoborockException("Got unexpected data type for user_data")
        return UserData.from_dict(user_data)

    async def pass_login(self, password: str) -> UserData:
        if not await self._login_limiter.try_acquire_async("login", blocking=True, timeout=1):
            raise RoborockRateLimit("Reached maximum requests for login. Please try again later.")
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()

        login_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
        login_response = await login_request.request(
            "post",
            "/api/v1/login",
            params={
                "username": self._username,
                "password": password,
                "needtwostepauth": "false",
            },
        )
        if login_response is None:
            raise RoborockException("Login response is none")
        if login_response.get("code") != 200:
            _LOGGER.info("Login failed for %s with the following context: %s", self._username, login_response)
            raise RoborockException(f"{login_response.get('msg')} - response code: {login_response.get('code')}")
        user_data = login_response.get("data")
        if not isinstance(user_data, dict):
            raise RoborockException("Got unexpected data type for user_data")
        return UserData.from_dict(user_data)

    async def pass_login_v3(self, password: str) -> UserData:
        """Seemingly it follows the format below, but password is encrypted in some manner.
        # login_response = await login_request.request(
        #     "post",
        #     "/api/v3/auth/email/login",
        #     params={
        #         "email": self._username,
        #         "password": password,
        #         "twoStep": 1,
        #         "version": 0
        #     },
        # )
        """
        raise NotImplementedError("Pass_login_v3 has not yet been implemented")

    async def code_login(self, code: int | str) -> UserData:
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()

        login_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
        login_response = await login_request.request(
            "post",
            "/api/v1/loginWithCode",
            params={
                "username": self._username,
                "verifycode": code,
                "verifycodetype": "AUTH_EMAIL_CODE",
            },
        )
        if login_response is None:
            raise RoborockException("Login request response is None")
        response_code = login_response.get("code")
        if response_code != 200:
            _LOGGER.info("Login failed for %s with the following context: %s", self._username, login_response)
            if response_code == 2018:
                raise RoborockInvalidCode("Invalid code - check your code and try again.")
            if response_code == 3009:
                raise RoborockNoUserAgreement("You must accept the user agreement in the Roborock app to continue.")
            if response_code == 3006:
                raise RoborockInvalidUserAgreement(
                    "User agreement must be accepted again - or you are attempting to use the Mi Home app account."
                )
            raise RoborockException(f"{login_response.get('msg')} - response code: {response_code}")
        user_data = login_response.get("data")
        if not isinstance(user_data, dict):
            raise RoborockException("Got unexpected data type for user_data")
        return UserData.from_dict(user_data)

    async def _get_home_id(self, user_data: UserData):
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()
        home_id_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
        home_id_response = await home_id_request.request(
            "get",
            "/api/v1/getHomeDetail",
            headers={"Authorization": user_data.token},
        )
        if home_id_response is None:
            raise RoborockException("home_id_response is None")
        if home_id_response.get("code") != 200:
            _LOGGER.info("Get Home Id failed with the following context: %s", home_id_response)
            if home_id_response.get("code") == 2010:
                raise RoborockInvalidCredentials(
                    f"Invalid credentials ({home_id_response.get('msg')}) - check your login and try again."
                )
            raise RoborockException(f"{home_id_response.get('msg')} - response code: {home_id_response.get('code')}")

        return home_id_response["data"]["rrHomeId"]

    async def get_home_data(self, user_data: UserData) -> HomeData:
        if not self._home_data_limiter.try_acquire("home_data", blocking=False):
            raise RoborockRateLimit("Reached maximum requests for home data. Please try again later.")
        rriot = user_data.rriot
        if rriot is None:
            raise RoborockException("rriot is none")
        home_id = await self._get_home_id(user_data)
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        home_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {
                "Authorization": _get_hawk_authentication(rriot, f"/user/homes/{str(home_id)}"),
            },
        )
        home_response = await home_request.request("get", "/user/homes/" + str(home_id))
        if not home_response.get("success"):
            raise RoborockException(home_response)
        home_data = home_response.get("result")
        if isinstance(home_data, dict):
            return HomeData.from_dict(home_data)
        else:
            raise RoborockException("home_response result was an unexpected type")

    async def get_home_data_v2(self, user_data: UserData) -> HomeData:
        """This is the same as get_home_data, but uses a different endpoint and includes non-robotic vacuums."""
        if not self._home_data_limiter.try_acquire("home_data", blocking=False):
            raise RoborockRateLimit("Reached maximum requests for home data. Please try again later.")
        rriot = user_data.rriot
        if rriot is None:
            raise RoborockException("rriot is none")
        home_id = await self._get_home_id(user_data)
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        home_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {
                "Authorization": _get_hawk_authentication(rriot, "/v2/user/homes/" + str(home_id)),
            },
        )
        home_response = await home_request.request("get", "/v2/user/homes/" + str(home_id))
        if not home_response.get("success"):
            raise RoborockException(home_response)
        home_data = home_response.get("result")
        if isinstance(home_data, dict):
            return HomeData.from_dict(home_data)
        else:
            raise RoborockException("home_response result was an unexpected type")

    async def get_home_data_v3(self, user_data: UserData) -> HomeData:
        """This is the same as get_home_data, but uses a different endpoint and includes non-robotic vacuums."""
        if not self._home_data_limiter.try_acquire("home_data", blocking=False):
            raise RoborockRateLimit("Reached maximum requests for home data. Please try again later.")
        rriot = user_data.rriot
        home_id = await self._get_home_id(user_data)
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        home_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {
                "Authorization": _get_hawk_authentication(rriot, "/v3/user/homes/" + str(home_id)),
            },
        )
        home_response = await home_request.request("get", "/v3/user/homes/" + str(home_id))
        if not home_response.get("success"):
            raise RoborockException(home_response)
        home_data = home_response.get("result")
        if isinstance(home_data, dict):
            return HomeData.from_dict(home_data)
        raise RoborockException(f"home_response result was an unexpected type: {home_data}")

    async def get_rooms(self, user_data: UserData, home_id: int | None = None) -> list[HomeDataRoom]:
        rriot = user_data.rriot
        if rriot is None:
            raise RoborockException("rriot is none")
        if home_id is None:
            home_id = await self._get_home_id(user_data)
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        room_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {
                "Authorization": _get_hawk_authentication(rriot, f"/user/homes/{home_id}/rooms"),
            },
        )
        room_response = await room_request.request("get", f"/user/homes/{home_id}/rooms")
        if not room_response.get("success"):
            raise RoborockException(room_response)
        rooms = room_response.get("result")
        if isinstance(rooms, list):
            output_list = []
            for room in rooms:
                output_list.append(HomeDataRoom.from_dict(room))
            return output_list
        else:
            raise RoborockException("home_response result was an unexpected type")

    async def get_shared_device_rooms(self, user_data: UserData, device_id: str) -> list[HomeDataRoom]:
        """Fetch room names for a shared (received) device."""
        rriot = user_data.rriot
        if rriot is None:
            raise RoborockException("rriot is none")
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        path = f"/user/deviceshare/query/{device_id}/rooms"
        room_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {"Authorization": _get_hawk_authentication(rriot, path)},
        )
        room_response = await room_request.request("get", path)
        if not room_response.get("success"):
            raise RoborockException(room_response)
        rooms = room_response.get("result")
        if isinstance(rooms, list):
            output_list = []
            for room in rooms:
                normalized_room = room
                if isinstance(room, dict) and "id" not in room and "roomId" in room:
                    normalized_room = {**room, "id": room["roomId"]}
                output_list.append(HomeDataRoom.from_dict(normalized_room))
            return output_list
        raise RoborockException("get_shared_device_rooms result was an unexpected type")

    async def get_scenes(self, user_data: UserData, device_id: str) -> list[HomeDataScene]:
        rriot = user_data.rriot
        if rriot is None:
            raise RoborockException("rriot is none")
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        scenes_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {
                "Authorization": _get_hawk_authentication(rriot, f"/user/scene/device/{str(device_id)}"),
            },
        )
        scenes_response = await scenes_request.request("get", f"/user/scene/device/{str(device_id)}")
        if not scenes_response.get("success"):
            raise RoborockException(scenes_response)
        scenes = scenes_response.get("result")
        if isinstance(scenes, list):
            return [HomeDataScene.from_dict(scene) for scene in scenes]
        else:
            raise RoborockException("scene_response result was an unexpected type")

    async def execute_scene(self, user_data: UserData, scene_id: int) -> None:
        rriot = user_data.rriot
        if rriot is None:
            raise RoborockException("rriot is none")
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        execute_scene_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {
                "Authorization": _get_hawk_authentication(rriot, f"/user/scene/{str(scene_id)}/execute"),
            },
        )
        execute_scene_response = await execute_scene_request.request("POST", f"/user/scene/{str(scene_id)}/execute")
        if not execute_scene_response.get("success"):
            raise RoborockException(execute_scene_response)

    async def get_schedules(self, user_data: UserData, device_id: str) -> list[HomeDataSchedule]:
        rriot = user_data.rriot
        if rriot is None:
            raise RoborockException("rriot is none")
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        schedules_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {
                "Authorization": _get_hawk_authentication(rriot, f"/user/devices/{device_id}/jobs"),
            },
        )
        schedules_response = await schedules_request.request("get", f"/user/devices/{str(device_id)}/jobs")
        if not schedules_response.get("success"):
            raise RoborockException(schedules_response)
        schedules = schedules_response.get("result")
        if isinstance(schedules, list):
            return [HomeDataSchedule.from_dict(schedule) for schedule in schedules]
        else:
            raise RoborockException(f"schedule_response result was an unexpected type: {schedules}")

    async def create_job(self, user_data: UserData, device_id: str, job: dict) -> dict:
        """Create a /jobs entry (schedule or one-time room clean) on a B01 device.

        Body-bearing writes must sign the request body in the Hawk payload slot and send those same
        compact bytes via ``data=``; ``json=`` would re-serialize with spaces and break the MAC.
        """
        rriot = user_data.rriot
        if rriot is None:
            raise RoborockException("rriot is none")
        if rriot.r.a is None:
            raise RoborockException("Missing field 'a' in rriot reference")
        path = f"/user/devices/{device_id}/jobs"
        job_request = PreparedRequest(
            rriot.r.a,
            self.session,
            {
                "Authorization": _get_hawk_authentication(rriot, path, body=job),
                "Content-Type": "application/json",
            },
        )
        response = await job_request.request("post", path, data=_compact_json(job).encode())
        if not response.get("success"):
            raise RoborockException(response)
        return response

    async def get_products(self, user_data: UserData) -> ProductResponse:
        """Gets all products and their schemas, good for determining status codes and model numbers."""
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()
        product_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
        product_response = await product_request.request(
            "get",
            "/api/v4/product",
            headers={"Authorization": user_data.token},
        )
        if product_response is None:
            raise RoborockException("home_id_response is None")
        if product_response.get("code") != 200:
            raise RoborockException(f"{product_response.get('msg')} - response code: {product_response.get('code')}")
        result = product_response.get("data")
        if isinstance(result, dict):
            return ProductResponse.from_dict(result)
        raise RoborockException("product result was an unexpected type")

    async def download_code(self, user_data: UserData, product_id: int):
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()
        product_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
        request = {"apilevel": 99999, "productids": [product_id], "type": 2}
        response = await product_request.request(
            "post",
            "/api/v1/appplugin",
            json=request,
            headers={"Authorization": user_data.token, "Content-Type": "application/json"},
        )
        return response["data"][0]["url"]

    async def download_category_code(self, user_data: UserData):
        base_url = await self.base_url
        header_clientid = self._get_header_client_id()
        product_request = PreparedRequest(base_url, self.session, {"header_clientid": header_clientid})
        response = await product_request.request(
            "get",
            "api/v1/plugins?apiLevel=99999&type=2",
            headers={
                "Authorization": user_data.token,
            },
        )
        return {r["category"]: r["url"] for r in response["data"]["categoryPluginList"]}


class PreparedRequest:
    def __init__(
        self, base_url: str, session: aiohttp.ClientSession | None = None, base_headers: dict | None = None
    ) -> None:
        self.base_url = base_url
        self.base_headers = base_headers or {}
        self.session = session

    async def request(self, method: str, url: str, params=None, data=None, headers=None, json=None) -> dict:
        _url = "/".join(s.strip("/") for s in [self.base_url, url])
        _headers = {**self.base_headers, **(headers or {})}
        close_session = self.session is None
        session = self.session if self.session is not None else aiohttp.ClientSession()
        try:
            async with session.request(method, _url, params=params, data=data, headers=_headers, json=json) as resp:
                return await resp.json()
        except ContentTypeError as err:
            """If we get an error, lets log everything for debugging."""
            try:
                resp_json = await resp.json(content_type=None)
                _LOGGER.info("Resp: %s", resp_json)
            except ContentTypeError as err_2:
                _LOGGER.info(err_2)
            resp_raw = await resp.read()
            _LOGGER.info("Resp raw: %s", resp_raw)
            # Still raise the err so that it's clear it failed.
            raise err
        except (aiohttp.ClientError, TimeoutError, OSError) as err:
            raise RoborockException(f"Network error contacting {_url}: {err}") from err
        finally:
            if close_session:
                await session.close()


def _process_extra_hawk_values(values: dict | None) -> str:
    if values is None:
        return ""
    else:
        sorted_keys = sorted(values.keys())
        result = []
        for key in sorted_keys:
            value = values.get(key)
            result.append(f"{key}={value}")
        return hashlib.md5("&".join(result).encode()).hexdigest()


def _compact_json(body: dict) -> str:
    """Serialize a JSON body to the exact compact bytes that are both signed and sent."""
    return json.dumps(body, separators=(",", ":"))


def _get_hawk_authentication(
    rriot: RRiot,
    url: str,
    formdata: dict | None = None,
    params: dict | None = None,
    body: dict | None = None,
) -> str:
    timestamp = math.floor(time.time())
    nonce = secrets.token_urlsafe(6)
    params_str = _process_extra_hawk_values(params)
    if body is not None:
        payload_str = hashlib.md5(_compact_json(body).encode()).hexdigest()
    else:
        payload_str = _process_extra_hawk_values(formdata)

    prestr = ":".join(
        [
            rriot.u,
            rriot.s,
            nonce,
            str(timestamp),
            hashlib.md5(url.encode()).hexdigest(),
            params_str,
            payload_str,
        ]
    )
    mac = base64.b64encode(hmac.new(rriot.h.encode(), prestr.encode(), hashlib.sha256).digest()).decode()
    return f'Hawk id="{rriot.u}",s="{rriot.s}",ts="{timestamp}",nonce="{nonce}",mac="{mac}"'


class UserWebApiClient:
    """Wrapper around RoborockApiClient to provide information for a specific user.

    This binds a RoborockApiClient to a specific user context with the
    provided UserData. This allows for easier access to user-specific data,
    to avoid needing to pass UserData around and mock out the web API.
    """

    def __init__(
        self, web_api: RoborockApiClient, user_data: UserData, unauthorized_hook: Callable[[], None] | None = None
    ) -> None:
        """Initialize the wrapper with the API client and user data."""
        self._web_api = web_api
        self._user_data = user_data
        self._unauthorized_hook = unauthorized_hook

    async def get_home_data(self) -> HomeData:
        """Fetch home data using the API client."""
        try:
            return await self._web_api.get_home_data_v3(self._user_data)
        except RoborockInvalidCredentials:
            if self._unauthorized_hook:
                self._unauthorized_hook()
            raise

    async def get_routines(self, device_id: str) -> list[HomeDataScene]:
        """Fetch routines (scenes) for a specific device."""
        try:
            return await self._web_api.get_scenes(self._user_data, device_id)
        except RoborockInvalidCredentials:
            if self._unauthorized_hook:
                self._unauthorized_hook()
            raise

    async def get_rooms(self) -> list[HomeDataRoom]:
        """Fetch rooms using the API client."""
        try:
            return await self._web_api.get_rooms(self._user_data)
        except RoborockInvalidCredentials:
            if self._unauthorized_hook:
                self._unauthorized_hook()
            raise

    async def get_shared_device_rooms(self, device_id: str) -> list[HomeDataRoom]:
        """Fetch shared-device rooms using the API client."""
        return await self._web_api.get_shared_device_rooms(self._user_data, device_id)

    async def execute_routine(self, scene_id: int) -> None:
        """Execute a specific routine (scene) by its ID."""
        try:
            await self._web_api.execute_scene(self._user_data, scene_id)
        except RoborockInvalidCredentials:
            if self._unauthorized_hook:
                self._unauthorized_hook()
            raise



from dataclasses import asdict

import pytest
from syrupy import SnapshotAssertion

from roborock import SHORT_MODEL_TO_ENUM
from roborock.data.code_mappings import RoborockProductNickname
from roborock.device_features import DeviceFeatures
from tests import mock_data


def test_supported_features_qrevo_maxv():
    """Ensure that a QREVO MaxV has some more complicated features enabled."""
    model = "roborock.vacuum.a87"
    product_nickname = SHORT_MODEL_TO_ENUM.get(model.split(".")[-1])
    device_features = DeviceFeatures.from_feature_flags(
        new_feature_info=4499197267967999,
        new_feature_info_str="508A977F7EFEFFFF",
        feature_info=[111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 121, 122, 123, 124, 125],
        product_nickname=product_nickname,
    )
    assert device_features
    print("\n".join(device_features.get_supported_features()))

    num_true = sum(v for v in vars(device_features).values() if isinstance(v, bool))
    print(num_true)
    assert num_true != 0
    assert device_features.is_dust_collection_setting_supported
    assert device_features.is_led_status_switch_supported
    assert not device_features.is_matter_supported
    print(device_features)


def test_supported_features_s7():
    """Ensure that a S7 has some more basic features enabled."""

    model = "roborock.vacuum.a15"
    product_nickname = SHORT_MODEL_TO_ENUM.get(model.split(".")[-1])
    device_features = DeviceFeatures.from_feature_flags(
        new_feature_info=636084721975295,
        new_feature_info_str="0000000000002000",
        feature_info=[111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 122, 123, 124, 125],
        product_nickname=product_nickname,
    )
    num_true = sum(v for v in vars(device_features).values() if isinstance(v, bool))
    assert num_true != 0
    assert device_features
    assert device_features.is_custom_mode_supported
    assert device_features.is_led_status_switch_supported
    assert not device_features.is_hot_wash_towel_supported
    num_true = sum(v for v in vars(device_features).values() if isinstance(v, bool))
    assert num_true != 0


def test_device_feature_serialization(snapshot: SnapshotAssertion) -> None:
    """Test serialization and deserialization of DeviceFeatures."""
    device_features = DeviceFeatures.from_feature_flags(
        new_feature_info=636084721975295,
        new_feature_info_str="0000000000002000",
        feature_info=[111, 112, 113, 114, 115, 116, 117, 118, 119, 120, 122, 123, 124, 125],
        product_nickname=RoborockProductNickname.TANOSS,
    )
    serialized = device_features.as_dict()
    deserialized = DeviceFeatures.from_dict(serialized)
    assert deserialized == device_features


def test_new_feature_str_missing():
    """Ensure that DeviceFeatures missing fields can still be created."""
    device_features = DeviceFeatures.from_feature_flags(
        new_feature_info=0,
        new_feature_info_str="",
        feature_info=[],
        product_nickname=None,
    )
    # Check arbitrary fields that are false by default
    assert not device_features.is_dust_collection_setting_supported
    assert not device_features.is_hot_wash_towel_supported
    assert not device_features.is_show_clean_finish_reason_supported


@pytest.mark.parametrize(
    ("device_filename"),
    list(mock_data.DEVICE_PRODUCT_PAIRS.keys()),
)
def test_device_features_from_home_data(
    device_filename: str,
    snapshot: SnapshotAssertion,
) -> None:
    """Test DeviceFeatures constructed from real testdata devices and products.

    For each paired device+product in testdata, construct DeviceFeatures from the
    featureSet/newFeatureSet home data fields and assert the full feature dict
    matches the snapshot. This catches regressions in feature-flag decoding
    across all real device samples.
    """
    device, product = mock_data.DEVICE_PRODUCT_PAIRS[device_filename]
    device_features = DeviceFeatures.from_feature_flags(
        new_feature_info=int(device.feature_set or "0"),
        new_feature_info_str=device.new_feature_set or "",
        feature_info=[],
        product_nickname=product.product_nickname,
    )
    feature_dict = {k: v for k, v in asdict(device_features).items() if isinstance(v, bool)}
    assert feature_dict == snapshot



import pytest

from roborock.data.b01_q10.b01_q10_code_mappings import (
    B01_Q10_DP,
    YXCleaningResult,
    YXCleanScope,
    YXCleanType,
    YXStartMethod,
)
from roborock.data.b01_q10.b01_q10_containers import Q10CleanRecord
from roborock.devices.traits.b01.q10 import Q10PropertiesApi
from roborock.devices.traits.b01.q10.clean_history import CleanHistoryTrait, CleanRecordConverter

from .conftest import FakeB01Q10Channel

# 12 underscore fields: recordId, startTime, cleanTime, cleanArea, mapLen, pathLen,
# virtualLen, cleanMode, workMode, cleaningResult, startMethod, collectDustCount.
RECORD_A = "abc123def456_1781226271_27_19_6692_12053_4_0_2_1_1_1"
RECORD_B = "def456abc789_1781139871_41_25_7100_18420_5_1_1_0_2_1"


@pytest.fixture(name="clean_history")
def clean_history_fixture(q10_api: Q10PropertiesApi) -> CleanHistoryTrait:
    return q10_api.clean_history


def _list_push(*records: str) -> dict[B01_Q10_DP, object]:
    """Build a decoded DPS update carrying a clean-record list."""
    return {B01_Q10_DP.CLEAN_RECORD: {"op": "list", "result": 1, "id": "", "data": list(records)}}


def _notify_push(record: str) -> dict[B01_Q10_DP, object]:
    """Build a decoded DPS update carrying a single clean-finished notify push."""
    return {B01_Q10_DP.CLEAN_RECORD: {"op": "notify", "result": 1, "id": record}}


def test_parse_clean_record() -> None:
    record = CleanRecordConverter.parse_record(RECORD_A)  # cleanMode/workMode/result/start = 0/2/1/1
    assert record is not None
    assert record.raw == RECORD_A
    assert record.record_id == "abc123def456"
    assert record.start_time == 1781226271
    assert record.clean_time == 27
    assert record.clean_area == 19
    assert record.map_len == 6692
    assert record.path_len == 12053
    assert record.virtual_len == 4
    assert record.collect_dust_count == 1
    assert record.start_datetime is not None
    # Enum-typed fields (no separate property converters).
    assert record.clean_mode is YXCleanScope.FULL
    assert record.work_mode is YXCleanType.VACUUM
    assert record.cleaning_result is YXCleaningResult.COMPLETED
    assert record.start_method is YXStartMethod.APP

    record_b = CleanRecordConverter.parse_record(RECORD_B)  # 1/1/0/2
    assert record_b is not None
    assert record_b.clean_mode is YXCleanScope.SELECTIVE_ROOM
    assert record_b.work_mode is YXCleanType.VAC_AND_MOP
    assert record_b.cleaning_result is YXCleaningResult.INTERRUPTED
    assert record_b.start_method is YXStartMethod.TIMER


@pytest.mark.parametrize(
    "raw",
    [
        "too_few_fields",
        "a_b_c_d_e_f_g_h_i_j_k_l_m",  # 13 fields
        "id_NOTANUMBER_27_19_6692_12053_4_0_2_1_1_1",  # non-int where int expected
    ],
)
def test_parse_clean_record_malformed_returns_none(raw: str) -> None:
    assert CleanRecordConverter.parse_record(raw) is None


def test_unmapped_enum_code_is_crash_safe() -> None:
    """An unmapped code (cleanMode 2) yields None on the field; the raw string is kept."""
    record = CleanRecordConverter.parse_record("x_1781226271_1_1_0_0_0_2_2_1_0_0")  # cleanMode=2, startMethod=0
    assert record is not None
    assert record.clean_mode is None  # code 2 unmapped -> None, never raises
    assert record.raw.split("_")[7] == "2"  # the original is retained in raw
    assert record.start_method is YXStartMethod.REMOTE  # 0 maps even though unobserved live
    assert Q10CleanRecord(raw="").clean_mode is None  # an unset field -> None


def test_update_from_dps_populates_records_newest_first(clean_history: CleanHistoryTrait) -> None:
    updates = 0

    def _on_update() -> None:
        nonlocal updates
        updates += 1

    clean_history.add_update_listener(_on_update)

    # RECORD_B starts earlier than RECORD_A; supply oldest-first to prove sorting.
    clean_history.update_from_dps(_list_push(RECORD_B, RECORD_A))

    assert [r.record_id for r in clean_history.records] == ["abc123def456", "def456abc789"]
    assert clean_history.last_record is not None
    assert clean_history.last_record.start_time == 1781226271
    assert updates == 1


def test_update_from_dps_skips_malformed_entries(clean_history: CleanHistoryTrait) -> None:
    clean_history.update_from_dps(_list_push(RECORD_A, "garbage", "1_2_3"))
    assert len(clean_history.records) == 1
    assert clean_history.records[0].record_id == "abc123def456"


@pytest.mark.parametrize(
    "dps",
    [
        {},  # no clean-record key at all
        {B01_Q10_DP.CLEAN_RECORD: {"op": "notify", "result": 1, "id": "x"}},  # malformed notify id
        {B01_Q10_DP.CLEAN_RECORD: False},  # bool sentinel, not a dict
    ],
)
def test_update_from_dps_ignores_unusable_pushes(
    clean_history: CleanHistoryTrait, dps: dict[B01_Q10_DP, object]
) -> None:
    updates = 0

    def _on_update() -> None:
        nonlocal updates
        updates += 1

    clean_history.add_update_listener(_on_update)
    clean_history.update_from_dps(dps)  # type: ignore[arg-type]

    assert clean_history.records == []
    assert clean_history.last_record is None
    assert updates == 0


def test_update_from_dps_notify_inserts_record(clean_history: CleanHistoryTrait) -> None:
    updates = 0

    def _on_update() -> None:
        nonlocal updates
        updates += 1

    clean_history.add_update_listener(_on_update)
    clean_history.update_from_dps(_notify_push(RECORD_A))

    assert [r.record_id for r in clean_history.records] == ["abc123def456"]
    assert updates == 1


def test_update_from_dps_notify_upserts_newest_first(clean_history: CleanHistoryTrait) -> None:
    clean_history.update_from_dps(_list_push(RECORD_B))  # older record present
    clean_history.update_from_dps(_notify_push(RECORD_A))  # newer -> prepend

    assert [r.record_id for r in clean_history.records] == ["abc123def456", "def456abc789"]

    # A notify for an existing record_id replaces it (no duplicate).
    updated_a = RECORD_A.replace("_27_", "_99_")  # same id, clean_time 27 -> 99
    clean_history.update_from_dps(_notify_push(updated_a))

    assert [r.record_id for r in clean_history.records] == ["abc123def456", "def456abc789"]
    assert clean_history.records[0].clean_time == 99


def test_update_from_dps_notify_ignores_malformed(clean_history: CleanHistoryTrait) -> None:
    clean_history.update_from_dps(_list_push(RECORD_A))
    clean_history.update_from_dps(_notify_push("garbage"))  # malformed id -> ignored

    assert [r.record_id for r in clean_history.records] == ["abc123def456"]


async def test_refresh_sends_op_list(q10_api: Q10PropertiesApi, fake_channel: FakeB01Q10Channel) -> None:
    await q10_api.clean_history.refresh()

    assert len(fake_channel.published_commands) == 1
    assert fake_channel.published_commands[0] == (
        B01_Q10_DP.COMMON,
        {"52": {"op": "list"}},
    )



"""Logging utilities for tests."""


class CapturedRequestLog:
    """Log of requests and responses for snapshot assertions.

    The log captures the raw bytes of each request and response along with
    a label indicating the direction of the message.
    """

    def __init__(self) -> None:
        """Initialize the request log."""
        self.entries: list[tuple[str, bytes]] = []

    def add_log_entry(self, label: str, data: bytes) -> None:
        """Add a request entry."""
        self.entries.append((label, data))

    def __repr__(self):
        """Return a string representation of the log entries.

        This assumes that the client will behave in a request-response manner,
        so each request is followed by a response. If a test uses non-deterministic
        message order, this may not be accurate and the test would need to decode
        the raw messages and remove any ordering assumptions.
        """
        lines = []
        for label, data in self.entries:
            lines.append(label)
            lines.extend(self._hexdump(data))
        return "\n".join(lines)

    def _hexdump(self, data: bytes, bytes_per_line: int = 16) -> list[str]:
        """Print a hexdump of the given bytes object in a tcpdump/hexdump -C style.

        This makes the packets easier to read and compare in test snapshots.

        Args:
            data: The bytes object to print.
            bytes_per_line: The number of bytes to display per line (default is 16).
        """

        # Use '.' for non-printable characters (ASCII < 32 or > 126)
        def to_printable_ascii(byte_val):
            return chr(byte_val) if 32 <= byte_val <= 126 else "."

        offset = 0
        lines = []
        while offset < len(data):
            chunk = data[offset : offset + bytes_per_line]
            # Format the hex values, space-padded to ensure alignment
            hex_values = " ".join(f"{byte:02x}" for byte in chunk)
            # Pad hex string to a fixed width so ASCII column lines up
            # 3 chars per byte ('xx ') for a full line of 16 bytes
            padded_hex = f"{hex_values:<{bytes_per_line * 3}}"
            # Format the ASCII values
            ascii_values = "".join(to_printable_ascii(byte) for byte in chunk)
            lines.append(f"{offset:08x}  {padded_hex} |{ascii_values}|")
            offset += bytes_per_line
        return lines



"""Tests for CleanSummaryTrait class for B01/Q7 devices."""

import json
import logging

import pytest

from roborock.data.b01_q7 import CleanRecordList
from roborock.devices.traits.b01.q7.clean_summary import CleanSummaryTrait
from roborock.exceptions import RoborockException

from .conftest import FakeQ7Channel

CLEAN_RECORD_LIST_DATA = {
    "total_time": 34980,
    "total_area": 28540,
    "total_count": 2,
    "record_list": [
        {
            "url": "/userdata/record_map/1766368207_1766368283_0_clean_map.bin",
            "detail": json.dumps(
                {
                    "record_start_time": 1766368207,
                    "method": 0,
                    "record_use_time": 60,
                    "clean_count": 1,
                    "record_clean_area": 85,
                    "record_clean_mode": 0,
                    "record_clean_way": 0,
                    "record_task_status": 20,
                    "record_faultcode": 0,
                    "record_dust_num": 0,
                    "clean_current_map": 0,
                    "record_map_url": "/userdata/record_map/1766368207_1766368283_0_clean_map.bin",
                }
            ),
        },
        {
            "url": "/userdata/record_map/1766369000_1766369200_0_clean_map.bin",
            "detail": json.dumps(
                {
                    "record_start_time": 1766369000,
                    "method": 1,
                    "record_use_time": 120,
                    "clean_count": 1,
                    "record_clean_area": 150,
                    "record_clean_mode": 1,
                    "record_clean_way": 0,
                    "record_task_status": 20,
                    "record_faultcode": 0,
                    "record_dust_num": 1,
                    "clean_current_map": 1,
                    "record_map_url": "/userdata/record_map/1766369000_1766369200_0_clean_map.bin",
                }
            ),
        },
    ],
}


@pytest.fixture(name="clean_summary_trait")
def clean_summary_trait_fixture(fake_channel: FakeQ7Channel) -> CleanSummaryTrait:
    return CleanSummaryTrait(fake_channel)


async def test_refresh_success(
    clean_summary_trait: CleanSummaryTrait,
    fake_channel: FakeQ7Channel,
) -> None:
    """Test successfully refreshing clean summary."""
    fake_channel.response_queue.append(CLEAN_RECORD_LIST_DATA)
    await clean_summary_trait.refresh()

    assert clean_summary_trait.total_time == 34980
    assert clean_summary_trait.total_area == 28540
    assert clean_summary_trait.total_count == 2
    assert clean_summary_trait.last_record_detail is not None
    assert clean_summary_trait.last_record_detail.record_start_time == 1766369000


async def test_refresh_with_no_records(
    clean_summary_trait: CleanSummaryTrait,
    fake_channel: FakeQ7Channel,
) -> None:
    """Test refreshing with no records."""
    empty_response = {
        "total_time": 0,
        "total_area": 0,
        "total_count": 0,
        "record_list": [],
    }
    fake_channel.response_queue.append(empty_response)
    await clean_summary_trait.refresh()

    assert clean_summary_trait.total_time == 0
    assert clean_summary_trait.total_area == 0
    assert clean_summary_trait.total_count == 0
    assert clean_summary_trait.last_record_detail is None


async def test_refresh_propagates_exceptions(
    clean_summary_trait: CleanSummaryTrait,
    fake_channel: FakeQ7Channel,
) -> None:
    """Test that exceptions from channel are propagated during refresh."""
    fake_channel.side_effect = RoborockException("Communication error")

    with pytest.raises(RoborockException, match="Communication error"):
        await clean_summary_trait.refresh()


async def test_get_clean_record_details_with_none_detail(
    clean_summary_trait: CleanSummaryTrait,
) -> None:
    """Test getting clean record details when some items have None detail."""
    response_with_none = {
        "total_time": 34980,
        "total_area": 28540,
        "total_count": 2,
        "record_list": [
            {
                "url": "/userdata/record_map/record1.bin",
                "detail": json.dumps(
                    {
                        "record_start_time": 1766368207,
                        "method": 0,
                        "record_use_time": 60,
                        "clean_count": 1,
                        "record_clean_area": 85,
                        "record_clean_mode": 0,
                        "record_clean_way": 0,
                        "record_task_status": 20,
                        "record_faultcode": 0,
                        "record_dust_num": 0,
                        "clean_current_map": 0,
                        "record_map_url": "/userdata/record_map/record1.bin",
                    }
                ),
            },
            {
                "url": "/userdata/record_map/record2.bin",
                "detail": None,
            },
        ],
    }

    details = await clean_summary_trait._get_clean_record_details(
        record_list=CleanRecordList.from_dict(response_with_none)
    )

    assert len(details) == 1
    assert details[0].record_start_time == 1766368207


async def test_get_clean_record_details_invalid_json(
    clean_summary_trait: CleanSummaryTrait,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that invalid JSON in detail is logged and skipped."""
    response_with_invalid_json = {
        "total_time": 34980,
        "total_area": 28540,
        "total_count": 1,
        "record_list": [
            {
                "url": "/userdata/record_map/record1.bin",
                "detail": "invalid json{",
            },
        ],
    }

    with caplog.at_level(logging.DEBUG):
        details = await clean_summary_trait._get_clean_record_details(
            record_list=CleanRecordList.from_dict(response_with_invalid_json)
        )

    assert len(details) == 0
    assert any("Failed to parse record detail" in record.message for record in caplog.records)



"""Diagnostics for debugging.

A Diagnostics object can be used to track counts and latencies of various
operations within a module. This can be useful for debugging performance issues
or understanding usage patterns.

This is an internal facing module and is not intended for public use. Diagnostics
data is collected and exposed to clients via higher level APIs like the
DeviceManager.
"""

import time
from collections import Counter
from collections.abc import Generator, Mapping
from contextlib import contextmanager
from typing import Any, Self, TypeVar, cast


class Diagnostics:
    """A class that holds diagnostics information for a module.

    You can use this class to hold counter or for recording timing information
    that can be exported as a dictionary for debugging purposes.
    """

    def __init__(self) -> None:
        """Initialize Diagnostics."""
        self._counter: Counter = Counter()
        self._subkeys: dict[str, Self] = {}

    def increment(self, key: str, count: int = 1) -> None:
        """Increment a counter for the specified key/event."""
        self._counter.update(Counter({key: count}))

    def elapsed(self, key_prefix: str, elapsed_ms: int = 1) -> None:
        """Track a latency event for the specified key/event prefix."""
        self.increment(f"{key_prefix}_count", 1)
        self.increment(f"{key_prefix}_sum", elapsed_ms)

    def as_dict(self) -> Mapping[str, Any]:
        """Return diagnostics as a debug dictionary."""
        data: dict[str, Any] = {k: self._counter[k] for k in self._counter}
        for k, d in self._subkeys.items():
            v = d.as_dict()
            if not v:
                continue
            data[k] = v
        return data

    def subkey(self, key: str) -> Self:
        """Return sub-Diagnostics object with the specified subkey.

        This will create a new Diagnostics object if one does not already exist
        for the specified subkey. Stats from the sub-Diagnostics will be included
        in the parent Diagnostics when exported as a dictionary.

        Args:
            key: The subkey for the diagnostics.

        Returns:
            The Diagnostics object for the specified subkey.
        """
        if key not in self._subkeys:
            self._subkeys[key] = type(self)()
        return self._subkeys[key]

    @contextmanager
    def timer(self, key_prefix: str) -> Generator[None, None, None]:
        """A context manager that records the timing of operations as a diagnostic."""
        start = time.perf_counter()
        try:
            yield
        finally:
            end = time.perf_counter()
            ms = int((end - start) * 1000)
            self.elapsed(key_prefix, ms)

    def reset(self) -> None:
        """Clear all diagnostics, for testing."""
        self._counter = Counter()
        for d in self._subkeys.values():
            d.reset()


T = TypeVar("T")

REDACT_KEYS = {
    # Potential identifiers
    "localKey",
    "mac",
    "bssid",
    "sn",
    "ip",
    "u",
    "s",
    "h",
    "k",
    # Large binary blobs are entirely omitted
    "imageContent",
    "mapData",
    "rawApiResponse",
    # Home data
    "id",  # We want to redact home_data.id but keep some other ids, see below
    "name",
    "productId",
    "ipAddress",
    "wifiName",
    "lat",
    "long",
}
KEEP_KEYS = {
    # Product information not unique per user
    "product.id",
    "product.schema.id",
    "product.schema.name",
    # Room ids are likely unique per user, but don't seem too sensitive and are
    # useful for debugging
    "rooms.id",
}
DEVICE_UID = "duid"
REDACTED = "**REDACTED**"


def redact_device_data(data: T, path: str = "") -> T | dict[str, Any]:
    """Redact sensitive data in a dict."""
    if not isinstance(data, (Mapping, list)):
        return data

    if isinstance(data, list):
        return cast(T, [redact_device_data(item, path) for item in data])

    redacted = {**data}

    for key, value in redacted.items():
        curr_path = f"{path}.{key}" if path else key
        if key in KEEP_KEYS or curr_path in KEEP_KEYS:
            continue
        if key in REDACT_KEYS or curr_path in REDACT_KEYS:
            redacted[key] = REDACTED
        elif key == DEVICE_UID and isinstance(value, str):
            redacted[key] = redact_device_uid(value)
        elif isinstance(value, dict):
            redacted[key] = redact_device_data(value, curr_path)
        elif isinstance(value, list):
            redacted[key] = [redact_device_data(item, curr_path) for item in value]

    return redacted


def redact_topic_name(topic: str) -> str:
    """Redact potentially identifying information from a topic name."""
    parts = topic.split("/")
    redacted_parts = parts[:4]
    for part in parts[4:]:
        if len(part) <= 5:
            redacted_parts.append("*****")
        else:
            redacted_parts.append("*****" + part[-5:])
    return "/".join(redacted_parts)


def redact_device_uid(duid: str) -> str:
    """Redact a device UID to hide identifying information."""
    return "******" + duid[-5:]



"""Tests for the V1Channel class.

This test simulates communication across both the MQTT and local connections
and failure modes, ensuring the V1Channel behaves correctly in various scenarios.
"""

import json
import logging
from collections.abc import Iterator
from unittest.mock import AsyncMock, Mock, patch

import pytest

from roborock.data import NetworkInfo, RoborockStateCode, S5MaxStatus, UserData
from roborock.devices.cache import DeviceCache, DeviceCacheData, InMemoryCache
from roborock.devices.rpc.v1_channel import V1Channel
from roborock.devices.transport.local_channel import LocalSession
from roborock.exceptions import RoborockException
from roborock.protocol import (
    create_local_decoder,
    create_local_encoder,
    create_mqtt_decoder,
    create_mqtt_encoder,
)
from roborock.protocols.v1_protocol import MapResponse, SecurityData, V1RpcChannel
from roborock.roborock_message import RoborockDataProtocol, RoborockMessage, RoborockMessageProtocol
from roborock.roborock_typing import RoborockCommand
from tests import mock_data
from tests.fixtures.channel_fixtures import FakeChannel

USER_DATA = UserData.from_dict(mock_data.USER_DATA)
TEST_DEVICE_UID = "abc123"
TEST_LOCAL_KEY = "local_key"
TEST_SECURITY_DATA = SecurityData(endpoint="test_endpoint", nonce=b"test_nonce_16byt")
TEST_HOST = mock_data.TEST_LOCAL_API_HOST


# Test messages for V1 protocol
TEST_REQUEST = RoborockMessage(
    protocol=RoborockMessageProtocol.RPC_REQUEST,
    payload=json.dumps({"dps": {"101": json.dumps({"id": 12346, "method": "get_status"})}}).encode(),
)
TEST_RESPONSE = RoborockMessage(
    protocol=RoborockMessageProtocol.RPC_RESPONSE,
    payload=json.dumps(
        {"dps": {"102": json.dumps({"id": 12346, "result": {"state": RoborockStateCode.cleaning}})}}
    ).encode(),
)
TEST_RESPONSE_2 = RoborockMessage(
    protocol=RoborockMessageProtocol.RPC_RESPONSE,
    payload=json.dumps(
        {"dps": {"102": json.dumps({"id": 12347, "result": {"state": RoborockStateCode.cleaning}})}}
    ).encode(),
)
TEST_NETWORK_INFO_RESPONSE = RoborockMessage(
    protocol=RoborockMessageProtocol.RPC_RESPONSE,
    payload=json.dumps({"dps": {"102": json.dumps({"id": 12345, "result": mock_data.NETWORK_INFO})}}).encode(),
)

TEST_NETWORKING_INFO = NetworkInfo.from_dict(mock_data.NETWORK_INFO)

# Encoders/Decoders
MQTT_ENCODER = create_mqtt_encoder(TEST_LOCAL_KEY)
MQTT_DECODER = create_mqtt_decoder(TEST_LOCAL_KEY)
LOCAL_ENCODER = create_local_encoder(TEST_LOCAL_KEY)
LOCAL_DECODER = create_local_decoder(TEST_LOCAL_KEY)


@pytest.fixture(name="mock_mqtt_channel")
async def setup_mock_mqtt_channel() -> FakeChannel:
    """Mock MQTT channel for testing."""
    channel = FakeChannel()
    await channel.connect()
    return channel


@pytest.fixture(name="mock_local_channel")
async def setup_mock_local_channel() -> FakeChannel:
    """Mock Local channel for testing."""
    return FakeChannel()


@pytest.fixture(name="mock_local_session")
def setup_mock_local_session(mock_local_channel: Mock) -> Mock:
    """Mock Local session factory for testing."""
    mock_session = Mock(spec=LocalSession)
    mock_session.return_value = mock_local_channel
    return mock_session


@pytest.fixture(name="mock_request_id", autouse=True)
def setup_mock_request_id() -> Iterator[None]:
    """Assign sequential request ids for testing."""

    next_id = 12345

    def fake_next_int(*args) -> int:
        nonlocal next_id
        id_to_return = next_id
        next_id += 1
        return id_to_return

    with patch("roborock.protocols.v1_protocol.get_next_int", side_effect=fake_next_int):
        yield


@pytest.fixture(name="mock_create_map_response_decoder")
def setup_mock_map_decoder() -> Iterator[Mock]:
    """Mock the map response decoder to control its behavior in tests."""
    with patch("roborock.devices.rpc.v1_channel.create_map_response_decoder") as mock_create_decoder:
        yield mock_create_decoder


@pytest.fixture(name="cache")
def cache_fixtures() -> InMemoryCache:
    """Mock cache for testing."""
    return InMemoryCache()


@pytest.fixture(name="device_cache")
def device_cache_fixtures() -> DeviceCache:
    """Mock device cache for testing."""
    return DeviceCache(TEST_DEVICE_UID, InMemoryCache())


@pytest.fixture(name="v1_channel")
def setup_v1_channel(
    mock_mqtt_channel: Mock,
    mock_local_session: Mock,
    mock_create_map_response_decoder: Mock,
    device_cache: DeviceCache,
) -> V1Channel:
    """Fixture to set up the V1 channel for tests."""
    return V1Channel(
        device_uid=TEST_DEVICE_UID,
        security_data=TEST_SECURITY_DATA,
        mqtt_channel=mock_mqtt_channel,
        local_session=mock_local_session,
        device_cache=device_cache,
    )


@pytest.fixture(name="rpc_channel")
def setup_rpc_channel(v1_channel: V1Channel) -> V1RpcChannel:
    """Fixture to set up the RPC channel for tests.

    We expect tests to use this to send commands via the V1Channel since we
    want to exercise the behavior that the V1RpcChannel is long lived and
    respects the current state of the underlying channels.
    """
    return v1_channel.rpc_channel


@pytest.fixture(name="mqtt_rpc_channel")
def setup_mqtt_rpc_channel(v1_channel: V1Channel) -> V1RpcChannel:
    """Fixture to set up the MQTT RPC channel for tests."""
    return v1_channel.mqtt_rpc_channel


@pytest.fixture(name="map_rpc_channel")
def setup_map_rpc_channel(v1_channel: V1Channel) -> V1RpcChannel:
    """Fixture to set up the Map RPC channel for tests."""
    return v1_channel.map_rpc_channel


@pytest.fixture(name="warning_caplog")
def setup_warning_caplog(caplog: pytest.LogCaptureFixture) -> pytest.LogCaptureFixture:
    """Fixture to capture warning messages."""
    caplog.set_level(logging.WARNING)
    return caplog


async def test_v1_channel_subscribe_mqtt_only_success(
    v1_channel: V1Channel,
    mock_mqtt_channel: FakeChannel,
    mock_local_session: Mock,
    mock_local_channel: FakeChannel,
) -> None:
    """Test successful subscription with MQTT only (local connection fails)."""
    # Setup: MQTT succeeds, local fails
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    mock_local_channel.connect.side_effect = RoborockException("Connection failed")

    callback = Mock()
    unsub = await v1_channel.subscribe(callback)

    # Verify MQTT connection was established
    assert mock_mqtt_channel.subscribers

    # Verify local connection was attempted but failed
    mock_local_session.assert_called_once_with(TEST_HOST)
    mock_local_channel.connect.assert_called_once()

    # Verify properties
    assert v1_channel.is_mqtt_connected
    assert not v1_channel.is_local_connected

    # Test unsubscribe
    unsub()
    assert not mock_mqtt_channel.subscribers


async def test_v1_channel_mqtt_disconnected(
    v1_channel: V1Channel,
    mock_mqtt_channel: FakeChannel,
    mock_local_session: Mock,
    mock_local_channel: FakeChannel,
) -> None:
    """Test successful subscription with MQTT only (local connection fails)."""
    # Setup: MQTT succeeds, local fails
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    mock_local_channel.connect.side_effect = RoborockException("Connection failed")

    callback = Mock()
    unsub = await v1_channel.subscribe(callback)

    # Verify MQTT connection was established
    assert mock_mqtt_channel.subscribers

    # Verify local connection was attempted but failed
    mock_local_session.assert_called_once_with(TEST_HOST)
    mock_local_channel.connect.assert_called_once()

    # Simulate an MQTT disconnection where the channel is not healthy
    mock_mqtt_channel.close()

    # Verify properties
    assert not v1_channel.is_mqtt_connected
    assert not v1_channel.is_local_connected

    # Test unsubscribe
    unsub()
    assert not mock_mqtt_channel.subscribers


async def test_v1_channel_mqtt_subscription_fails_local_succeeds(
    v1_channel: V1Channel,
    mock_mqtt_channel: FakeChannel,
    device_cache: DeviceCache,
) -> None:
    """Test MQTT subscription failure while local connection succeeds."""
    # Pre-populate cache so we don't query network info via MQTT
    device_cache_data = await device_cache.get()
    device_cache_data.network_info = TEST_NETWORKING_INFO
    await device_cache.set(device_cache_data)

    # Simulate MQTT subscription failing
    mock_mqtt_channel.subscribe.side_effect = RoborockException("MQTT subscription failed")

    # Subscribe should succeed via local fallback
    callback = Mock()
    unsub = await v1_channel.subscribe(callback)

    # Verify MQTT is not reported as connected, but local is
    assert not v1_channel.is_mqtt_connected
    assert v1_channel.is_local_connected
    assert v1_channel.is_connected

    unsub()


async def test_v1_channel_all_connection_attempts_fail(
    v1_channel: V1Channel,
    mock_mqtt_channel: FakeChannel,
    mock_local_channel: FakeChannel,
    device_cache: DeviceCache,
) -> None:
    """Test when both local connect and MQTT subscribe fail."""
    # Pre-populate cache so we don't query network info via MQTT
    device_cache_data = await device_cache.get()
    device_cache_data.network_info = TEST_NETWORKING_INFO
    await device_cache.set(device_cache_data)

    mock_local_channel.connect.side_effect = RoborockException("local down")
    mock_mqtt_channel.subscribe.side_effect = RoborockException("MQTT subscription failed")

    with pytest.raises(RoborockException):
        await v1_channel.subscribe(Mock())

    # After a failed subscription, properties should reflect no active connection
    assert not v1_channel.is_mqtt_connected
    assert not v1_channel.is_local_connected
    assert not v1_channel.is_connected


async def test_v1_channel_subscribe_local_success(
    v1_channel: V1Channel,
    mock_mqtt_channel: Mock,
    mock_local_channel: Mock,
    mock_local_session: Mock,
) -> None:
    """Test successful subscription with local connections."""
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)

    # Mock network info retrieval
    callback = Mock()
    unsub = await v1_channel.subscribe(callback)

    # Verify local connection was attempted and succeeded
    mock_local_session.assert_called_once_with(TEST_HOST)
    mock_local_channel.connect.assert_called_once()

    # Verify mqtt is also established
    assert mock_mqtt_channel.subscribers
    assert mock_local_channel.subscribers

    # Verify properties
    assert v1_channel.is_mqtt_connected
    assert v1_channel.is_local_connected

    # Test unsubscribe cleans up both
    unsub()
    assert not mock_mqtt_channel.subscribers
    assert not mock_local_channel.subscribers


async def test_v1_channel_subscribe_already_connected_error(v1_channel: V1Channel, mock_mqtt_channel: Mock) -> None:
    """Test error when trying to subscribe when already connected."""
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)

    # First subscription succeeds
    await v1_channel.subscribe(Mock())

    # Second subscription should fail
    with pytest.raises(ValueError, match="Only one subscription allowed at a time"):
        await v1_channel.subscribe(Mock())


async def test_v1_channel_send_command_local_preferred(
    v1_channel: V1Channel,
    mock_mqtt_channel: Mock,
    mock_local_channel: Mock,
    rpc_channel: V1RpcChannel,
) -> None:
    """Test command sending prefers local connection when available."""
    # Establish connections
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    await v1_channel.subscribe(Mock())

    # Send command
    mock_local_channel.response_queue.append(TEST_RESPONSE)
    result = await rpc_channel.send_command(
        RoborockCommand.GET_STATUS,
        response_type=S5MaxStatus,
    )

    # Verify local response was parsed
    assert result.state == RoborockStateCode.cleaning


async def test_v1_channel_send_command_local_fails(
    v1_channel: V1Channel,
    mock_mqtt_channel: Mock,
    mock_local_channel: Mock,
    rpc_channel: V1RpcChannel,
) -> None:
    """Test case where sending with local connection fails, falling back to MQTT."""

    # Establish connections
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    await v1_channel.subscribe(Mock())

    # Local command fails
    mock_local_channel.publish = Mock()
    mock_local_channel.publish.side_effect = RoborockException("Local failed")

    # MQTT command succeeds
    mock_mqtt_channel.response_queue.append(TEST_RESPONSE)

    # Send command
    result = await rpc_channel.send_command(
        RoborockCommand.GET_STATUS,
        response_type=S5MaxStatus,
    )

    # Verify result
    assert result.state == RoborockStateCode.cleaning

    # Verify local was attempted
    mock_local_channel.publish.assert_called_once()

    # Verify MQTT was used
    assert mock_mqtt_channel.published_messages
    # The last message should be the command we sent
    assert mock_mqtt_channel.published_messages[-1].protocol == RoborockMessageProtocol.RPC_REQUEST


@pytest.mark.parametrize(
    ("local_channel_side_effect", "local_channel_responses", "mock_mqtt_channel_responses"),
    [
        (RoborockException("Local failed"), [], [TEST_RESPONSE]),
        (None, [], [TEST_RESPONSE]),
        (None, [RoborockMessage(protocol=RoborockMessageProtocol.RPC_RESPONSE, payload=b"invalid")], [TEST_RESPONSE]),
    ],
    ids=[
        "local-fails-mqtt-succeeds",
        "local-no-response-mqtt-succeeds",
        "local-invalid-response-mqtt-succeeds",
    ],
)
async def test_v1_channel_send_pick_first_available(
    v1_channel: V1Channel,
    rpc_channel: V1RpcChannel,
    mock_mqtt_channel: Mock,
    mock_local_channel: Mock,
    local_channel_side_effect: Exception | None,
    local_channel_responses: list[RoborockMessage],
    mock_mqtt_channel_responses: list[RoborockMessage],
) -> None:
    """Test command sending works with MQTT only."""
    # Setup: only MQTT connection
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    mock_local_channel.connect.side_effect = local_channel_side_effect

    await v1_channel.subscribe(Mock())

    # Send command
    mock_mqtt_channel.response_queue.extend(mock_mqtt_channel_responses)
    mock_local_channel.response_queue.extend(local_channel_responses)
    result = await rpc_channel.send_command(
        RoborockCommand.GET_STATUS,
        response_type=S5MaxStatus,
    )

    # Verify only MQTT was used
    assert result.state == RoborockStateCode.cleaning


async def test_v1_channel_send_decoded_command_with_params(
    v1_channel: V1Channel,
    rpc_channel: V1RpcChannel,
    mock_mqtt_channel: Mock,
    mock_local_channel: Mock,
) -> None:
    """Test command sending with parameters."""

    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    await v1_channel.subscribe(Mock())

    # Send command with params
    mock_local_channel.response_queue.append(TEST_RESPONSE)
    test_params = {"volume": 80}
    await rpc_channel.send_command(
        RoborockCommand.CHANGE_SOUND_VOLUME,
        response_type=S5MaxStatus,
        params=test_params,
    )

    # Verify command was sent with correct params
    assert mock_local_channel.published_messages
    sent_message = mock_local_channel.published_messages[0]
    assert sent_message
    assert isinstance(sent_message, RoborockMessage)
    assert sent_message.payload
    payload = sent_message.payload.decode()
    json_data = json.loads(payload)
    assert "dps" in json_data
    assert "101" in json_data["dps"]
    decoded_payload = json.loads(json_data["dps"]["101"])
    assert decoded_payload["method"] == "change_sound_volume"
    assert decoded_payload["params"] == {"volume": 80}


async def test_v1_channel_networking_info_retrieved_during_connection(
    v1_channel: V1Channel,
    mock_mqtt_channel: Mock,
    mock_local_channel: Mock,
    mock_local_session: Mock,
) -> None:
    """Test that networking information is retrieved during local connection setup."""
    # Setup: MQTT returns network info when requested
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)

    # Subscribe - this should trigger network info retrieval for local connection
    await v1_channel.subscribe(Mock())

    # Verify local connection was esablished
    assert v1_channel.is_local_connected

    # Verify network info was requested via MQTT
    assert mock_mqtt_channel.published_messages

    # Verify local session was created with the correct IP
    mock_local_session.assert_called_once_with(mock_data.NETWORK_INFO["ip"])


async def test_v1_channel_networking_info_cached_during_connection(
    mock_mqtt_channel: Mock,
    mock_local_channel: Mock,
    mock_local_session: Mock,
) -> None:
    """Test that networking information is cached and reused on subsequent connections."""

    # Create a cache with pre-populated network info
    device_cache_data = DeviceCacheData()
    device_cache_data.network_info = TEST_NETWORKING_INFO

    mock_device_cache = AsyncMock()
    mock_device_cache.get.return_value = device_cache_data
    mock_device_cache.set = AsyncMock()

    # Create V1Channel with the mock cache
    v1_channel = V1Channel(
        device_uid=TEST_DEVICE_UID,
        security_data=TEST_SECURITY_DATA,
        mqtt_channel=mock_mqtt_channel,
        local_session=mock_local_session,
        device_cache=mock_device_cache,
    )

    # Subscribe - should use cached network info
    await v1_channel.subscribe(Mock())

    # Verify local connections are established
    assert v1_channel.is_local_connected

    # Verify network info was NOT requested via MQTT (cache hit)
    assert not mock_mqtt_channel.published_messages
    assert not mock_local_channel.published_messages

    # Verify local session was created with the correct IP from cache
    mock_local_session.assert_called_once_with(mock_data.NETWORK_INFO["ip"])

    # Verify cache was accessed but not updated (cache hit)
    mock_device_cache.get.assert_called_once()
    mock_device_cache.set.assert_not_called()


# V1Channel edge cases tests


async def test_v1_channel_local_connect_network_info_failure(
    v1_channel: V1Channel,
    mock_mqtt_channel: Mock,
) -> None:
    """Test local connection when network info retrieval fails."""
    mock_mqtt_channel.publish_side_effect = RoborockException("Network info failed")

    with pytest.raises(RoborockException):
        await v1_channel._local_connect()


async def test_v1_channel_local_connect_network_info_failure_fallback_to_cache(
    mock_mqtt_channel: FakeChannel,
    mock_local_session: Mock,
    v1_channel: V1Channel,
    device_cache: DeviceCache,
) -> None:
    """Test local connection falls back to cache when network info retrieval fails."""
    # Create a cache with pre-populated network info
    await device_cache.set(DeviceCacheData(network_info=TEST_NETWORKING_INFO))

    # Setup: MQTT fails to publish
    mock_mqtt_channel.publish_side_effect = RoborockException("Network info failed")

    # Attempt local connect, forcing a refresh (prefer_cache=False)
    # This should try MQTT, fail, and then fall back to cache
    await v1_channel._local_connect(prefer_cache=False)

    # Verify local connection was established
    assert v1_channel.is_local_connected

    # Verify MQTT was attempted (published message)
    assert mock_mqtt_channel.published_messages

    # Verify local session was created with the correct IP from cache
    mock_local_session.assert_called_once_with(TEST_HOST)


async def test_v1_channel_command_encoding_validation(
    v1_channel: V1Channel,
    mqtt_rpc_channel: V1RpcChannel,
    rpc_channel: V1RpcChannel,
    mock_mqtt_channel: Mock,
    mock_local_channel: Mock,
) -> None:
    """Test that command encoding works for different protocols."""
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    await v1_channel.subscribe(Mock())

    # Send mqtt command and capture the request
    mock_mqtt_channel.response_queue.append(TEST_RESPONSE)
    await mqtt_rpc_channel.send_command(RoborockCommand.CHANGE_SOUND_VOLUME, params={"volume": 50})
    assert mock_mqtt_channel.published_messages
    mqtt_message = mock_mqtt_channel.published_messages[0]

    # Send local command and capture the request
    mock_local_channel.response_queue.append(TEST_RESPONSE_2)
    await rpc_channel.send_command(RoborockCommand.CHANGE_SOUND_VOLUME, params={"volume": 50})
    assert mock_local_channel.published_messages
    local_message = mock_local_channel.published_messages[0]

    # Verify both are RoborockMessage instances
    assert isinstance(mqtt_message, RoborockMessage)
    assert isinstance(local_message, RoborockMessage)

    # But they should have different protocols
    assert mqtt_message.protocol == RoborockMessageProtocol.RPC_REQUEST
    assert local_message.protocol == RoborockMessageProtocol.GENERAL_REQUEST


async def test_v1_channel_send_map_command(
    v1_channel: V1Channel,
    map_rpc_channel: V1RpcChannel,
    mock_mqtt_channel: Mock,
    mock_create_map_response_decoder: Mock,
) -> None:
    """Test that the map channel can correctly decode a map response."""
    # Establish connections
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    await v1_channel.subscribe(Mock())

    # Prepare a mock map response
    decompressed_map_data = b"this is the decompressed map data"
    request_id = 12346  # from the mock_request_id fixture

    # Mock the decoder to return a known response
    map_response = MapResponse(request_id=request_id, data=decompressed_map_data)
    mock_create_map_response_decoder.return_value.return_value = map_response

    # The actual message content doesn't matter as much since the decoder is mocked
    map_response_message = RoborockMessage(
        protocol=RoborockMessageProtocol.MAP_RESPONSE,
        payload=b"dummy_payload",
    )
    mock_mqtt_channel.response_queue.append(map_response_message)

    # Send the command and get the result
    result = await map_rpc_channel.send_command(RoborockCommand.GET_MAP_V1)

    # Verify the result is the data from our mocked decoder
    assert result == decompressed_map_data


async def test_v1_channel_add_dps_listener(
    v1_channel: V1Channel,
    mock_mqtt_channel: FakeChannel,
) -> None:
    """Test that DPS listeners receive decoded protocol updates from MQTT."""
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    await v1_channel.subscribe(Mock())

    # Create a mock listener for DPS updates
    dps_listener = Mock()
    unsub_dps = v1_channel.add_dps_listener(dps_listener)

    # Simulate an incoming MQTT message with data protocol payload.
    dps_payload = json.dumps({"dps": {"121": 5}}).encode()
    push_message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_REQUEST,
        payload=dps_payload,
    )
    mock_mqtt_channel.notify_subscribers(push_message)

    dps_listener.assert_called_once()
    called_args = dps_listener.call_args[0][0]
    assert called_args[RoborockDataProtocol.STATE] == 5

    unsub_dps()

    # Verify unsubscribe works
    dps_listener.reset_mock()
    v1_channel._on_mqtt_message(push_message)
    dps_listener.assert_not_called()


async def test_v1_channel_dps_listener_raises_exception(
    v1_channel: V1Channel,
    mock_mqtt_channel: FakeChannel,
) -> None:
    """Test that DPS listener that raises exceptions is ignored."""
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    await v1_channel.subscribe(Mock())

    # Create a mock listener for DPS updates
    dps_listener1 = Mock()
    dps_listener1.side_effect = Exception("DPS listener failed")
    dps_listener2 = Mock()
    unsub_dps1 = v1_channel.add_dps_listener(dps_listener1)
    unsub_dps2 = v1_channel.add_dps_listener(dps_listener2)

    # Simulate an incoming MQTT message with data protocol payload.
    dps_payload = json.dumps({"dps": {"121": 5}}).encode()
    push_message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_REQUEST,
        payload=dps_payload,
    )
    mock_mqtt_channel.notify_subscribers(push_message)

    dps_listener1.assert_called_once()
    dps_listener2.assert_called_once()

    unsub_dps1()
    unsub_dps2()


async def test_v1_channel_resubscribe_after_unsub(
    v1_channel: V1Channel,
    mock_mqtt_channel: FakeChannel,
) -> None:
    """A subscribe -> unsub -> subscribe cycle must not raise, and the new callback works.

    Regression: unsub() previously failed to reset the subscription, so the second
    subscribe() tripped the "Only one subscription allowed at a time" guard.
    This is the exact failure that bricked a second vacuum sharing an account.
    """
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    callback = Mock()
    unsub = await v1_channel.subscribe(callback)

    # The subscribed callback receives messages arriving on the channel.
    mock_mqtt_channel.notify_subscribers(TEST_RESPONSE)
    callback.assert_called_once_with(TEST_RESPONSE)

    # After unsub, the old callback no longer receives messages.
    unsub()
    callback.reset_mock()
    mock_mqtt_channel.notify_subscribers(TEST_RESPONSE)
    callback.assert_not_called()

    # Re-subscribing must succeed (network info is now cached, no MQTT needed) and
    # the new callback then receives messages.
    new_callback = Mock()
    unsub2 = await v1_channel.subscribe(new_callback)
    mock_mqtt_channel.notify_subscribers(TEST_RESPONSE)
    new_callback.assert_called_once_with(TEST_RESPONSE)
    unsub2()


async def test_v1_channel_subscribe_failure_is_atomic(
    v1_channel: V1Channel,
    mock_mqtt_channel: FakeChannel,
    mock_local_channel: FakeChannel,
) -> None:
    """A failure partway through subscribe() leaves the channel re-subscribable.

    Regression: a failed subscribe() previously leaked the background reconnect
    task and a partial subscription, so the next attempt could neither reuse nor
    cleanly recreate the channel.
    """
    # Both transports down: local connect fails and the MQTT subscribe fails.
    mock_local_channel.connect.side_effect = RoborockException("local down")
    mock_mqtt_channel.subscribe.side_effect = RoborockException("mqtt down")

    with pytest.raises(RoborockException):
        await v1_channel.subscribe(Mock())

    # The failed subscribe left no dangling subscription on the channel.
    assert not mock_mqtt_channel.subscribers

    # And the channel is re-subscribable once the transports recover: the new
    # subscription succeeds and its callback receives messages.
    mock_local_channel.connect.side_effect = None
    mock_mqtt_channel.subscribe.side_effect = mock_mqtt_channel._subscribe
    mock_mqtt_channel.response_queue.append(TEST_NETWORK_INFO_RESPONSE)
    callback = Mock()
    unsub = await v1_channel.subscribe(callback)
    mock_mqtt_channel.notify_subscribers(TEST_RESPONSE)
    callback.assert_called_once_with(TEST_RESPONSE)
    unsub()



import base64
import gzip
import hashlib
import io
import zlib
from pathlib import Path

import pytest
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad
from PIL import Image

from roborock.exceptions import RoborockException
from roborock.map.b01_map_parser import B01MapParser, _parse_scmap_payload
from roborock.map.proto.b01_scmap_pb2 import RobotMap  # type: ignore[attr-defined]
from roborock.protocols.b01_q7_protocol import create_map_key, decode_map_payload

FIXTURE = Path(__file__).resolve().parent / "testdata" / "raw-mqtt-map301.bin.inflated.bin.gz"


def _derive_map_key(serial: str, model: str) -> bytes:
    model_suffix = model.split(".")[-1]
    model_key = (model_suffix + "0" * 16)[:16].encode()
    material = f"{serial}+{model_suffix}+{serial}".encode()
    encrypted = AES.new(model_key, AES.MODE_ECB).encrypt(pad(material, AES.block_size))
    md5 = hashlib.md5(base64.b64encode(encrypted), usedforsecurity=False).hexdigest()
    return md5[8:24].encode()


def test_b01_map_parser_decodes_and_renders_fixture() -> None:
    serial = "testsn012345"
    model = "roborock.vacuum.sc05"
    inflated = gzip.decompress(FIXTURE.read_bytes())

    compressed = zlib.compress(inflated)
    map_key = create_map_key(serial, model)
    encrypted = AES.new(map_key.key, AES.MODE_ECB).encrypt(pad(compressed.hex().encode(), AES.block_size))
    payload = base64.b64encode(encrypted)

    parser = B01MapParser()
    inflated_payload = decode_map_payload(payload, map_key=map_key)
    parsed = parser.parse(inflated_payload)

    assert parsed.image_content is not None
    assert parsed.image_content.startswith(b"\x89PNG\r\n\x1a\n")
    assert parsed.map_data is not None

    # The fixture includes 10 rooms with names room1..room10.
    assert parsed.map_data.additional_parameters["room_names"] == {
        10: "room1",
        11: "room2",
        12: "room3",
        13: "room4",
        14: "room5",
        15: "room6",
        16: "room7",
        17: "room8",
        18: "room9",
        19: "room10",
    }

    # Image should be scaled by default.
    img = Image.open(io.BytesIO(parsed.image_content))
    assert img.size == (340 * 4, 300 * 4)


def test_b01_scmap_parser_maps_observed_schema_fields() -> None:
    payload = RobotMap()
    payload.mapType = 1
    payload.mapExtInfo.taskBeginDate = 100
    payload.mapExtInfo.mapUploadDate = 200
    payload.mapExtInfo.mapValid = 1
    payload.mapExtInfo.mapVersion = 3
    payload.mapExtInfo.boudaryInfo.mapMd5 = "md5"
    payload.mapExtInfo.boudaryInfo.vMinX = 10
    payload.mapExtInfo.boudaryInfo.vMaxX = 20
    payload.mapExtInfo.boudaryInfo.vMinY = 30
    payload.mapExtInfo.boudaryInfo.vMaxY = 40
    payload.mapHead.mapHeadId = 7
    payload.mapHead.sizeX = 2
    payload.mapHead.sizeY = 2
    payload.mapHead.minX = 1.5
    payload.mapHead.minY = 2.5
    payload.mapHead.maxX = 3.5
    payload.mapHead.maxY = 4.5
    payload.mapHead.resolution = 0.05
    payload.mapData.mapData = bytes([0, 127, 128, 128])

    room_one = payload.roomDataInfo.add()
    room_one.roomId = 42
    room_one.roomName = "Kitchen"
    room_one.cleanState = 1
    room_one.roomNamePost.x = 11.25
    room_one.roomNamePost.y = 22.5
    room_one.colorId = 7
    room_one.global_seq = 9

    room_two = payload.roomDataInfo.add()
    room_two.roomId = 99
    room_two.cleanState = 0

    parsed = _parse_scmap_payload(payload.SerializeToString())

    assert parsed.mapType == 1
    assert parsed.HasField("mapExtInfo")
    assert parsed.mapExtInfo.taskBeginDate == 100
    assert parsed.mapExtInfo.mapUploadDate == 200
    assert parsed.mapExtInfo.HasField("boudaryInfo")
    assert parsed.mapExtInfo.boudaryInfo.vMaxY == 40
    assert parsed.HasField("mapHead")
    assert parsed.mapHead.mapHeadId == 7
    assert parsed.mapHead.sizeX == 2
    assert parsed.mapHead.sizeY == 2
    assert parsed.mapHead.resolution == pytest.approx(0.05)
    assert parsed.HasField("mapData")
    assert parsed.mapData.HasField("mapData")
    assert parsed.mapData.mapData == bytes([0, 127, 128, 128])
    assert parsed.roomDataInfo[0].roomId == 42
    assert parsed.roomDataInfo[0].roomName == "Kitchen"
    assert parsed.roomDataInfo[0].HasField("roomNamePost")
    assert parsed.roomDataInfo[0].roomNamePost.x == pytest.approx(11.25)
    assert parsed.roomDataInfo[0].roomNamePost.y == pytest.approx(22.5)
    assert parsed.roomDataInfo[0].colorId == 7
    assert parsed.roomDataInfo[0].global_seq == 9
    assert parsed.roomDataInfo[1].roomId == 99
    assert not parsed.roomDataInfo[1].HasField("roomName")


def test_b01_map_parser_rejects_invalid_payload() -> None:
    parser = B01MapParser()
    with pytest.raises(RoborockException, match="Failed to parse B01 SCMap"):
        parser.parse(b"not a map")



"""Decoders for Q10 (B01/ss07) map vector overlays.

No-go zones, no-mop zones, virtual walls and zoned-clean areas are not part of
the map raster; the device reports them as base64-encoded blobs in separate data
points (``dpRestrictedZoneUp`` 55, ``dpVirtualWallUp`` 57, ``dpZonedUp`` 59).

The blob format was reverse-engineered from a live ss07 (confirmed against 7
real no-go zones):

    [version: u8][count: u8] then ``count`` fixed-size records, each:
        [type: u8][vertex_count: u8] then vertex_count (x, y) int16-BE pairs,
        zero-padded to the record size.

Coordinates are in the device's world units (the same space as the cleaning
path), so a :class:`~roborock.map.b01_grid_layers.GridCalibration` maps them to
map pixels. ``type`` distinguishes the restriction kind (2 = no-mop, 3 = door
threshold, anything else -- incl. 0 -- a no-go zone); it is preserved verbatim
so callers can route polygons to the right ``MapData`` layer.
"""

import base64
from dataclasses import dataclass, field

_DEFAULT_RECORD_SIZE = 38  # 2-byte record header + up to 9 (x, y) int16 pairs


@dataclass
class Q10Zone:
    """A polygon overlay (no-go / no-mop / virtual wall) in world coordinates."""

    type: int
    vertices: list[tuple[int, int]] = field(default_factory=list)


def _as_bytes(data: bytes | str | None) -> bytes:
    if data is None:
        return b""
    if isinstance(data, bytes):
        return data
    try:
        return base64.b64decode(data + "=" * (-len(data) % 4))
    except (ValueError, base64.binascii.Error):  # type: ignore[attr-defined]
        return b""


def parse_zone_blob(data: bytes | str | None) -> list[Q10Zone]:
    """Decode a Q10 zone/wall overlay blob into a list of :class:`Q10Zone`.

    Accepts the raw bytes or the base64 string straight from the data point.
    Returns ``[]`` for empty/absent/unparsable blobs (the device sends a single
    ``0x00`` byte when there are none).
    """
    raw = _as_bytes(data)
    if len(raw) < 2:
        return []
    count = raw[1]
    if count <= 0:
        return []

    body = raw[2:]
    record_size = len(body) // count if count and len(body) % count == 0 else _DEFAULT_RECORD_SIZE
    if record_size < 2:
        return []

    zones: list[Q10Zone] = []
    for index in range(count):
        record = body[index * record_size : (index + 1) * record_size]
        if len(record) < 2:
            break
        zone_type = record[0]
        vertex_count = record[1]
        needed = 2 + vertex_count * 4
        if needed > len(record):
            continue  # malformed record; skip rather than misread padding
        vertices = [
            (
                int.from_bytes(record[2 + j * 4 : 4 + j * 4], "big", signed=True),
                int.from_bytes(record[4 + j * 4 : 6 + j * 4], "big", signed=True),
            )
            for j in range(vertex_count)
        ]
        zones.append(Q10Zone(type=zone_type, vertices=vertices))
    return zones


# Observed ``type`` values, confirmed against an ss07 Q10 (firmware 03.11.24) and
# cross-checked with the ioBroker roborock adapter: 2 = no-mop, 3 = door
# threshold, 1 = virtual wall. Any other value (including 0) is a no-go zone.
# In practice virtual walls arrive on a separate DP (VIRTUAL_WALL_UP 57), so this
# restricted-zone DP normally only carries 0 / 2 / 3. The raw value is also kept
# on ``Q10Zone.type`` for callers that recognise it.
#
# Corrected from an earlier reading that treated type 3 as no-mop -- 3 is the
# door-threshold rectangle; the no-mop area reads back as type 2. Reported and
# verified by @andrewlyeats (ss07 read-backs + the ioBroker Q10 parser).
ZONE_TYPE_NO_GO = 0
ZONE_TYPE_VIRTUAL_WALL = 1
ZONE_TYPE_NO_MOP = 2
ZONE_TYPE_THRESHOLD = 3



"""End-to-end tests for MQTT session.

These tests use a fake MQTT broker to verify the session implementation. We
mock out the lower level socket connections to simulate a broker which gets us
close to an "end to end" test without needing an actual MQTT broker server.

These are higher level tests than the similar tests in tests/mqtt/test_roborock_session.py
which use mocks to verify specific behaviors.
"""

import asyncio
import json
from collections.abc import AsyncGenerator, Awaitable, Callable
from typing import Any

import pytest
import syrupy
from Crypto.Cipher import AES
from Crypto.Util.Padding import pad

from roborock.data.b01_q7 import WorkStatusMapping
from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP
from roborock.data.containers import UserData
from roborock.data.zeo.zeo_code_mappings import ZeoState
from roborock.devices.cache import Cache, InMemoryCache
from roborock.devices.device_manager import DeviceManager, UserParams, create_device_manager
from roborock.protocol import MessageParser
from roborock.protocols.a01_protocol import A01_VERSION
from roborock.protocols.b01_q7_protocol import B01_VERSION
from roborock.protocols.v1_protocol import LocalProtocolVersion
from roborock.roborock_message import RoborockB01Props, RoborockMessage, RoborockMessageProtocol, RoborockZeoProtocol
from roborock.web_api import RoborockApiClient
from tests import mock_data, mqtt_packet
from tests.fixtures.logging import CapturedRequestLog
from tests.mock_data import HOME_DATA_RAW, LOCAL_KEY

TEST_USERNAME = "user@example.com"
TEST_CODE = 1234

# The topic used for the user + device. This is determined from the fake Home
# data API response.
TEST_TOPIC_FORMAT = "rr/m/o/user123/19648f94/{duid}"
TEST_RANDOM = 23
TEST_HOST = mock_data.TEST_LOCAL_API_HOST
NETWORK_INFO = {
    "ip": TEST_HOST,
    "ssid": "test_wifi",
    "mac": "aa:bb:cc:dd:ee:ff",
    "bssid": "aa:bb:cc:dd:ee:ff",
    "rssi": -50,
}
# For tests that want to skip the web API login flow
TEST_USER_PARAMS = UserParams(
    username=TEST_USERNAME,
    user_data=UserData.from_dict(mock_data.USER_DATA),
    base_url=mock_data.BASE_URL,
)
MQTT_DEFAULT_RESPONSES: list[bytes] = [
    # MQTT connection response
    mqtt_packet.gen_connack(rc=0, flags=2),
    # ACK the request to subscribe to the topic
    mqtt_packet.gen_suback(mid=1),
]


@pytest.fixture(autouse=True)
def auto_mock_mqtt_client(mock_aiomqtt_client: None) -> None:
    """Automatically use the mock mqtt client fixture."""


@pytest.fixture(autouse=True)
def auto_fast_backoff(fast_backoff_fixture: None) -> None:
    """Automatically use the fast backoff fixture."""


@pytest.fixture(autouse=True)
def mqtt_server_fixture(mock_paho_mqtt_create_connection: None, mock_paho_mqtt_select: None) -> None:
    """Fixture to mock the MQTT connection.

    This is here to pull in the mock socket fixtures into all tests used here.
    """


@pytest.fixture(autouse=True)
def auto_mock_local_client(mock_async_create_local_connection: None) -> None:
    """Automatically use the mock local client fixture."""


@pytest.fixture(name="device_manager_factory")
async def device_manager_factory_fixture() -> AsyncGenerator[Callable[[UserParams], Awaitable[DeviceManager]], None]:
    """Fixture to create a device manager and handle auto shutdown on test failure."""

    cleanup_tasks: list[Callable[[], Awaitable[None]]] = []
    cache: Cache = InMemoryCache()

    async def factory(user_params: UserParams) -> DeviceManager:
        """Create a device manager and auto cleanup."""
        device_manager = await create_device_manager(user_params, cache=cache)
        cleanup_tasks.append(device_manager.close)
        return device_manager

    yield factory

    await asyncio.gather(*[task() for task in cleanup_tasks])


class ResponseBuilder:
    """Utility class to build raw response messages.

    This helps keep track of sequence numbers and timestamps mostly to remove
    them from the main test body. These are mostly ignored by the client in the
    response.
    """

    def __init__(self) -> None:
        """Initialize the response builder."""
        self.seq_counter = 0
        self.timestamp_counter = 1766520441
        self.connect_nonce: int | None = None
        self.ack_nonce: int | None = None
        self.protocol = RoborockMessageProtocol.RPC_RESPONSE
        self.version: bytes = LocalProtocolVersion.V1.value.encode()

    def build_v1(
        self,
        payload: bytes,
        protocol: RoborockMessageProtocol | None = None,
    ) -> bytes:
        """Build an encoded response message."""
        self.seq_counter += 1
        return self._encrypt(
            self._build_roborock_message(
                payload=payload,
                protocol=protocol if protocol is not None else self.protocol,
            ),
        )

    def build_v1_rpc(
        self,
        data: dict[str, Any],
    ) -> bytes:
        """Build an encoded RPC response message."""
        self.timestamp_counter += 1
        return self.build_v1(
            payload=json.dumps(
                {
                    "t": self.timestamp_counter,
                    "dps": {
                        "102": json.dumps(data),
                    },
                }
            ).encode(),
        )

    def build_a01_rpc(self, data: dict[str, Any]) -> bytes:
        """Build an encoded A01 RPC response message."""
        self.timestamp_counter += 1
        return self._encrypt(
            self._build_roborock_message(
                payload=pad(json.dumps({"dps": data}).encode(), AES.block_size),
            ),
        )

    def build_b01_q7_rpc(self, data: dict[str, Any] | str, code: int | None = None, msg_id: int | None = None) -> bytes:
        """Build an encoded B01 RPC response message."""
        message: dict[str, Any] = {
            "msgId": str(msg_id),
            "data": data,
        }
        if code is not None:
            message["code"] = code
        return self._build_b01_dps(message)

    def _build_b01_dps(self, message: dict[str, Any] | str) -> bytes:
        """Build an encoded B01 RPC response message given an inner message."""
        dps_payload = {"dps": {"10000": json.dumps(message)}}
        self.seq_counter += 1
        return self._encrypt(
            self._build_roborock_message(
                payload=json.dumps(dps_payload).encode(),
            ),
        )

    def _build_roborock_message(
        self,
        payload: bytes,
        protocol: RoborockMessageProtocol | None = None,
    ) -> RoborockMessage:
        """Build a Roborock message."""
        return RoborockMessage(
            protocol=protocol if protocol is not None else self.protocol,
            random=TEST_RANDOM,
            seq=self.seq_counter,
            payload=payload,
            version=self.version,
        )

    def _encrypt(self, message: RoborockMessage) -> bytes:
        """Encrypt a message."""
        return MessageParser.build(
            message,
            local_key=LOCAL_KEY,
            connect_nonce=self.connect_nonce,
            ack_nonce=self.ack_nonce,
        )


async def test_v1_device(
    mock_rest: Any,
    push_mqtt_response: Callable[[bytes], None],
    local_response_queue: asyncio.Queue[bytes],
    local_received_requests: asyncio.Queue[bytes],
    log: CapturedRequestLog,
    snapshot: syrupy.SnapshotAssertion,
    device_manager_factory: Callable[[UserParams], Awaitable[DeviceManager]],
) -> None:
    """Test the device manager end to end flow with a v1 device."""

    # Simulate the login flow to get user params
    web_api = RoborockApiClient(username=TEST_USERNAME)
    await web_api.request_code()
    user_data = await web_api.code_login(TEST_CODE)

    # Prepare MQTT requests
    response_builder = ResponseBuilder()
    test_topic = TEST_TOPIC_FORMAT.format(duid="abc123")
    mqtt_responses: list[bytes] = [
        *MQTT_DEFAULT_RESPONSES,
        # ACK the GET_NETWORK_INFO call. id is deterministic based on deterministic_message_fixtures
        mqtt_packet.gen_publish(
            test_topic, mid=2, payload=response_builder.build_v1_rpc(data={"id": 9090, "result": NETWORK_INFO})
        ),
    ]
    for response in mqtt_responses:
        push_mqtt_response(response)

    # Prepare local device responses. The ids are deterministic based on deterministic_message_fixtures
    response_builder.seq_counter = 0
    local_responses: list[bytes] = [
        # Queue HELLO response
        response_builder.build_v1(protocol=RoborockMessageProtocol.HELLO_RESPONSE, payload=b"ok"),
        # Feature discovery part 1 & 2
        response_builder.build_v1_rpc(data={"id": 9094, "result": [mock_data.APP_GET_INIT_STATUS]}),
        response_builder.build_v1_rpc(data={"id": 9097, "result": [mock_data.STATUS]}),
    ]
    for payload in local_responses:
        local_response_queue.put_nowait(payload)

    # Create the device manager
    user_params = UserParams(
        username=TEST_USERNAME,
        user_data=user_data,
        base_url=await web_api.base_url,
    )
    device_manager = await device_manager_factory(user_params)

    # The mocked Home Data API returns a single v1 device
    devices = await device_manager.get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.duid == "abc123"
    assert device.name == "Roborock S7 MaxV"
    assert device.is_connected
    assert device.is_local_connected

    # Verify GET_STATUS response based on mock_data.STATUS
    assert device.v1_properties
    assert device.v1_properties.status
    assert device.v1_properties.status.state_name == "charging"
    assert device.v1_properties.status.battery == 100
    assert device.v1_properties.status.clean_time == 1176

    # Verify arbitrary device features
    assert device.v1_properties.device_features.is_show_clean_finish_reason_supported
    assert device.v1_properties.device_features.is_customized_clean_supported
    assert not device.v1_properties.device_features.is_matter_supported

    # Close the device manager. We will test re-connecting and reusing the network
    # information and device discovery information from the cache.
    await device_manager.close()

    mqtt_responses = [
        *MQTT_DEFAULT_RESPONSES,
        # No network info call this time since it should be cached
    ]
    for response in mqtt_responses:
        push_mqtt_response(response)

    # Prepare local device responses.
    response_builder.seq_counter = 0
    local_response_queue.put_nowait(
        response_builder.build_v1(protocol=RoborockMessageProtocol.HELLO_RESPONSE, payload=b"ok")
    )

    device_manager = await device_manager_factory(user_params)

    # The mocked Home Data API returns a single v1 device
    devices = await device_manager.get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.duid == "abc123"
    assert device.name == "Roborock S7 MaxV"
    assert device.is_connected
    assert device.is_local_connected

    # Verify arbitrary device features from cache
    assert device.v1_properties
    assert device.v1_properties.device_features
    assert device.v1_properties.device_features.is_show_clean_finish_reason_supported
    assert device.v1_properties.device_features.is_customized_clean_supported
    assert not device.v1_properties.device_features.is_matter_supported

    # In the previous test, the dock information is fetched and has the side effect of
    # populating the status trait. This test gets dock information from the cache so
    # we have to manually refresh status the first time (like other traits).
    assert device.v1_properties
    assert device.v1_properties.status
    assert device.v1_properties.status.state_name is None

    # Exercise a GET_STATUS call. id is deterministic based on deterministic_message_fixtures
    local_response_queue.put_nowait(response_builder.build_v1_rpc(data={"id": 9101, "result": [mock_data.STATUS]}))

    # Verify GET_STATUS response
    await device.v1_properties.status.refresh()
    assert device.v1_properties.status.state_name == "charging"

    assert snapshot == log


async def test_l01_device(
    mock_rest: Any,
    push_mqtt_response: Callable[[bytes], None],
    local_response_queue: asyncio.Queue[bytes],
    local_received_requests: asyncio.Queue[bytes],
    log: CapturedRequestLog,
    snapshot: syrupy.SnapshotAssertion,
    device_manager_factory: Callable[[UserParams], Awaitable[DeviceManager]],
) -> None:
    """Test the device manager end to end flow with a l01 device."""
    # Prepare MQTT requests
    mqtt_response_builder = ResponseBuilder()
    test_topic = TEST_TOPIC_FORMAT.format(duid="abc123")
    mqtt_responses: list[bytes] = [
        *MQTT_DEFAULT_RESPONSES,
        # ACK the GET_NETWORK_INFO call. id is deterministic based on deterministic_message_fixtures
        mqtt_packet.gen_publish(
            test_topic, mid=2, payload=mqtt_response_builder.build_v1_rpc(data={"id": 9090, "result": NETWORK_INFO})
        ),
    ]
    for response in mqtt_responses:
        push_mqtt_response(response)

    # Prepare local device responses. The ids are deterministic based on deterministic_message_fixtures
    local_response_builder = ResponseBuilder()
    local_response_builder.version = LocalProtocolVersion.L01.value.encode()
    local_response_builder.connect_nonce = 9093
    local_responses: list[bytes] = [
        # Initial V1.0 Hello request will fail and cause a retry with L01
        b"\x00",
        # Queue HELLO response with L01
        local_response_builder.build_v1(protocol=RoborockMessageProtocol.HELLO_RESPONSE, payload=b"ok"),
    ]
    # Feature discovery requests are sent with an ack nonce based on the random sent in HELLO_RESPONSE
    local_response_builder.ack_nonce = TEST_RANDOM
    local_responses.extend(
        [
            local_response_builder.build_v1_rpc(data={"id": 9094, "result": [mock_data.APP_GET_INIT_STATUS]}),
            local_response_builder.build_v1_rpc(data={"id": 9097, "result": [mock_data.STATUS]}),
        ]
    )
    for payload in local_responses:
        local_response_queue.put_nowait(payload)

    # Create the device manager
    device_manager = await device_manager_factory(TEST_USER_PARAMS)

    # The mocked Home Data API returns a single v1 device
    devices = await device_manager.get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.duid == "abc123"
    assert device.name == "Roborock S7 MaxV"
    assert device.is_connected
    assert device.is_local_connected

    # Verify GET_STATUS response based on mock_data.STATUS
    assert device.v1_properties
    assert device.v1_properties.status
    assert device.v1_properties.status.state_name == "charging"
    assert device.v1_properties.status.battery == 100
    assert device.v1_properties.status.clean_time == 1176

    # Verify arbitrary device features
    assert device.v1_properties.device_features.is_show_clean_finish_reason_supported
    assert device.v1_properties.device_features.is_customized_clean_supported
    assert not device.v1_properties.device_features.is_matter_supported

    assert snapshot == log


@pytest.mark.parametrize(
    "home_data",
    [
        (
            {
                **HOME_DATA_RAW,
                "devices": [mock_data.Q10_DEVICE_DATA],
                "products": [mock_data.SS07_PRODUCT_DATA],
            }
        )
    ],
)
async def test_q10_device(
    mock_rest: Any,
    push_mqtt_response: Callable[[bytes], None],
    log: CapturedRequestLog,
    device_manager_factory: Callable[[UserParams], Awaitable[DeviceManager]],
    home_data: dict[str, Any],
    snapshot: syrupy.SnapshotAssertion,
) -> None:
    """Test the device manager end to end flow with a B01 Q10 device."""
    # Prepare MQTT requests
    for response in MQTT_DEFAULT_RESPONSES:
        push_mqtt_response(response)

    # Create the device manager
    device_manager = await device_manager_factory(TEST_USER_PARAMS)

    # The mocked Home Data API returns a single v1 device
    devices = await device_manager.get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.duid == "device-id-def456"
    assert device.name == "Roborock Q10 S5+"
    assert device.is_connected
    assert not device.is_local_connected  # Q10 does not support local connections

    # Send a command. We don't block any response, but just use this to verify
    # against the golden byte stream snapshot.
    assert device.b01_q10_properties
    command = device.b01_q10_properties.command
    await command.send(B01_Q10_DP.REQUEST_DPS, params={})

    # In the future here we can verify receiving requests from the device

    assert snapshot == log


@pytest.mark.parametrize(
    "home_data",
    [
        (
            {
                **HOME_DATA_RAW,
                "devices": [mock_data.Q7_DEVICE_DATA],
                "products": [mock_data.SC01_PRODUCT_DATA],
            }
        )
    ],
)
async def test_q7_device(
    mock_rest: Any,
    push_mqtt_response: Callable[[bytes], None],
    log: CapturedRequestLog,
    device_manager_factory: Callable[[UserParams], Awaitable[DeviceManager]],
    home_data: dict[str, Any],
    snapshot: syrupy.SnapshotAssertion,
) -> None:
    """Test the device manager end to end flow with a B01 Q10 device."""
    # Prepare MQTT requests
    response_builder = ResponseBuilder()
    response_builder.version = B01_VERSION
    test_topic = TEST_TOPIC_FORMAT.format(duid="device-id-q7")
    mqtt_responses: list[bytes] = [
        *MQTT_DEFAULT_RESPONSES,
        # ACK the Query status call sent below. id is deterministic based on deterministic_message_fixtures
        mqtt_packet.gen_publish(
            test_topic, mid=2, payload=response_builder.build_b01_q7_rpc({"status": 2}, msg_id=9090)
        ),
        # ACK the start clean call sent below. id is deterministic based on deterministic_message_fixtures
        mqtt_packet.gen_publish(test_topic, mid=2, payload=response_builder.build_b01_q7_rpc("ok", msg_id=9093)),
    ]
    for response in mqtt_responses:
        push_mqtt_response(response)

    # Create the device manager
    device_manager = await device_manager_factory(TEST_USER_PARAMS)

    # The mocked Home Data API returns a single v1 device
    devices = await device_manager.get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.duid == "device-id-q7"
    assert device.name == "Roborock Q7"
    assert device.is_connected
    assert not device.is_local_connected  # Q7 does not support local connections

    # Query a value from the device.
    assert device.b01_q7_properties
    props = await device.b01_q7_properties.query_values([RoborockB01Props.STATUS])
    assert props
    assert props.status == WorkStatusMapping.PAUSED

    # Send a command and block on an OK response.
    await device.b01_q7_properties.start_clean()

    assert snapshot == log


@pytest.mark.parametrize(
    "home_data",
    [
        (
            {
                **HOME_DATA_RAW,
                "devices": [mock_data.ZEO_ONE_DEVICE_DATA],
                "products": [mock_data.A102_PRODUCT_DATA],
            }
        )
    ],
)
async def test_a01_device(
    mock_rest: Any,
    push_mqtt_response: Callable[[bytes], None],
    log: CapturedRequestLog,
    device_manager_factory: Callable[[UserParams], Awaitable[DeviceManager]],
    home_data: dict[str, Any],
    snapshot: syrupy.SnapshotAssertion,
) -> None:
    """Test the device manager end to end flow with an A01 device."""
    # Prepare MQTT requests
    response_builder = ResponseBuilder()
    response_builder.version = A01_VERSION
    test_topic = TEST_TOPIC_FORMAT.format(duid="zeo_duid")
    mqtt_responses: list[bytes] = [
        *MQTT_DEFAULT_RESPONSES,
        # ACK the Query state call sent below. id is deterministic based on deterministic_message_fixtures
        mqtt_packet.gen_publish(test_topic, mid=2, payload=response_builder.build_a01_rpc({"203": 6})),
    ]
    for response in mqtt_responses:
        push_mqtt_response(response)

    # Create the device manager
    device_manager = await device_manager_factory(TEST_USER_PARAMS)

    # The mocked Home Data API returns a single v1 device
    devices = await device_manager.get_devices()
    assert len(devices) == 1
    device = devices[0]
    assert device.duid == "zeo_duid"
    assert device.name == "Zeo One"
    assert device.is_connected
    assert not device.is_local_connected  # Washing Machine does not support local connections

    # Query a value from the device.
    assert device.zeo
    props: dict[RoborockZeoProtocol, Any] = await device.zeo.query_values([RoborockZeoProtocol.STATE])
    assert props
    assert props[RoborockZeoProtocol.STATE] == ZeoState.spinning.name

    assert snapshot == log



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: roborock/map/proto/b01_scmap.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'roborock/map/proto/b01_scmap.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"roborock/map/proto/b01_scmap.proto\x12\tb01.scmap\"\'\n\x0f\x44\x65vicePointInfo\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\"]\n\x0fMapBoundaryInfo\x12\x0e\n\x06mapMd5\x18\x01 \x01(\t\x12\r\n\x05vMinX\x18\x02 \x01(\r\x12\r\n\x05vMaxX\x18\x03 \x01(\r\x12\r\n\x05vMinY\x18\x04 \x01(\r\x12\r\n\x05vMaxY\x18\x05 \x01(\r\"\xd9\x01\n\nMapExtInfo\x12\x15\n\rtaskBeginDate\x18\x01 \x01(\r\x12\x15\n\rmapUploadDate\x18\x02 \x01(\r\x12\x10\n\x08mapValid\x18\x03 \x01(\r\x12\x0e\n\x06radian\x18\x04 \x01(\r\x12\r\n\x05\x66orce\x18\x05 \x01(\r\x12\x11\n\tcleanPath\x18\x06 \x01(\r\x12/\n\x0b\x62oudaryInfo\x18\x07 \x01(\x0b\x32\x1a.b01.scmap.MapBoundaryInfo\x12\x12\n\nmapVersion\x18\x08 \x01(\r\x12\x14\n\x0cmapValueType\x18\t \x01(\r\"\x8a\x01\n\x0bMapHeadInfo\x12\x11\n\tmapHeadId\x18\x01 \x01(\r\x12\r\n\x05sizeX\x18\x02 \x01(\r\x12\r\n\x05sizeY\x18\x03 \x01(\r\x12\x0c\n\x04minX\x18\x04 \x01(\x02\x12\x0c\n\x04minY\x18\x05 \x01(\x02\x12\x0c\n\x04maxX\x18\x06 \x01(\x02\x12\x0c\n\x04maxY\x18\x07 \x01(\x02\x12\x12\n\nresolution\x18\x08 \x01(\x02\"\x1e\n\x0bMapDataInfo\x12\x0f\n\x07mapData\x18\x01 \x01(\x0c\"\x87\x02\n\x0cRoomDataInfo\x12\x0e\n\x06roomId\x18\x01 \x01(\r\x12\x10\n\x08roomName\x18\x02 \x01(\t\x12\x12\n\nroomTypeId\x18\x03 \x01(\r\x12\x12\n\nmeterialId\x18\x04 \x01(\r\x12\x12\n\ncleanState\x18\x05 \x01(\r\x12\x11\n\troomClean\x18\x06 \x01(\r\x12\x16\n\x0eroomCleanIndex\x18\x07 \x01(\r\x12\x30\n\x0croomNamePost\x18\x08 \x01(\x0b\x32\x1a.b01.scmap.DevicePointInfo\x12\x0f\n\x07\x63olorId\x18\n \x01(\r\x12\x17\n\x0f\x66loor_direction\x18\x0b \x01(\r\x12\x12\n\nglobal_seq\x18\x0c \x01(\r\"\xc7\x01\n\x08RobotMap\x12\x0f\n\x07mapType\x18\x01 \x01(\r\x12)\n\nmapExtInfo\x18\x02 \x01(\x0b\x32\x15.b01.scmap.MapExtInfo\x12\'\n\x07mapHead\x18\x03 \x01(\x0b\x32\x16.b01.scmap.MapHeadInfo\x12\'\n\x07mapData\x18\x04 \x01(\x0b\x32\x16.b01.scmap.MapDataInfo\x12-\n\x0croomDataInfo\x18\x0c \x03(\x0b\x32\x17.b01.scmap.RoomDataInfo')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'roborock.map.proto.b01_scmap_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_DEVICEPOINTINFO']._serialized_start=49
  _globals['_DEVICEPOINTINFO']._serialized_end=88
  _globals['_MAPBOUNDARYINFO']._serialized_start=90
  _globals['_MAPBOUNDARYINFO']._serialized_end=183
  _globals['_MAPEXTINFO']._serialized_start=186
  _globals['_MAPEXTINFO']._serialized_end=403
  _globals['_MAPHEADINFO']._serialized_start=406
  _globals['_MAPHEADINFO']._serialized_end=544
  _globals['_MAPDATAINFO']._serialized_start=546
  _globals['_MAPDATAINFO']._serialized_end=576
  _globals['_ROOMDATAINFO']._serialized_start=579
  _globals['_ROOMDATAINFO']._serialized_end=842
  _globals['_ROBOTMAP']._serialized_start=845
  _globals['_ROBOTMAP']._serialized_end=1044
# @@protoc_insertion_point(module_scope)



from enum import StrEnum
from typing import Self

from ..code_mappings import RoborockEnum


class RoborockFinishReason(RoborockEnum):
    manual_interrupt = 21  # Cleaning interrupted by user
    cleanup_interrupted = 24  # Cleanup interrupted
    manual_interrupt_2 = 21
    manual_interrupt_12 = 29
    breakpoint = 32  # Could not continue cleaning
    breakpoint_2 = 33
    cleanup_interrupted_2 = 34
    manual_interrupt_3 = 35
    manual_interrupt_4 = 36
    manual_interrupt_5 = 37
    manual_interrupt_6 = 43
    locate_fail = 45  # Positioning Failed
    cleanup_interrupted_3 = 64
    locate_fail_2 = 65
    manual_interrupt_7 = 48
    manual_interrupt_8 = 49
    manual_interrupt_9 = 50
    cleanup_interrupted_4 = 51
    finished_cleaning = 52  # Finished cleaning
    finished_cleaning_2 = 54
    finished_cleaning_3 = 55
    finished_cleaning_4 = 56
    finished_clenaing_5 = 57
    manual_interrupt_10 = 60
    area_unreachable = 61  # Area unreachable
    area_unreachable_2 = 62
    washing_error = 67  # Washing error
    back_to_wash_failure = 68  # Failed to return to the dock
    cleanup_interrupted_5 = 101
    breakpoint_4 = 102
    manual_interrupt_11 = 103
    cleanup_interrupted_6 = 104
    cleanup_interrupted_7 = 105
    cleanup_interrupted_8 = 106
    cleanup_interrupted_9 = 107
    cleanup_interrupted_10 = 109
    cleanup_interrupted_11 = 110
    patrol_success = 114  # Cruise completed
    patrol_fail = 115  # Cruise failed
    pet_patrol_success = 116  # Pet found
    pet_patrol_fail = 117  # Pet found failed


class RoborockInCleaning(RoborockEnum):
    complete = 0
    global_clean_not_complete = 1
    zone_clean_not_complete = 2
    segment_clean_not_complete = 3


class RoborockCleanType(RoborockEnum):
    all_zone = 1
    draw_zone = 2
    select_zone = 3
    quick_build = 4
    video_patrol = 5
    pet_patrol = 6


class RoborockChargeStatus(RoborockEnum):
    """Describes the charging status of the device."""

    unknown = -1
    charge_waiting = 0
    charging = 1


class RoborockDockState(StrEnum):
    """Synthesized high-level dock and power state of the device.

    This enum represents a unified "UI-level" state that combines multiple raw
    device data points (`state`, `charge_status`, `battery`) into a single,
    human-readable status that accurately reflects what the vacuum is doing
    relative to the dock.

    It is highly recommended for consumers of this API
    to use this synthesized state to determine if the vacuum is charging or
    docked, rather than attempting to parse the raw integer data points, as
    this safely handles backward compatibility for older models that lack
    explicit off-peak schedule reporting.
    """

    unknown = "unknown"
    """The dock state could not be determined or is unmapped."""

    idle = "idle"
    """The vacuum is away from the dock (e.g., cleaning, paused, or errored).
    In the official app, this state presents the 'Return to Dock' or 'Recharge' action."""

    returning = "returning"
    """The vacuum is actively navigating its way back to the dock.
    In the official app, this state presents the 'Stop' or 'Pause' action."""

    charging = "charging"
    """The vacuum is on the dock and actively receiving electricity.
    In the official app, this state is displayed as 'Charging'."""

    off_peak_waiting = "off_peak_waiting"
    """The vacuum is on the dock but charging is paused. It is waiting for the
    user's scheduled 'Valley Electricity' off-peak hours to begin before
    drawing power.
    In the official app, this state is displayed as 'Charging paused during peak hours'."""

    full = "full"
    """The vacuum is on the dock and the battery is at 100% capacity.
    In the official app, this state is displayed as 'Fully charged'."""

    dusting = "dusting"
    """The vacuum is on the dock and is currently being evacuated by the
    auto-empty base.
    In the official app, this state is displayed as 'Emptying dustbin'."""


class RoborockStartType(RoborockEnum):
    button = 1
    app = 2
    schedule = 3
    mi_home = 4
    quick_start = 5
    voice_control = 13
    routines = 101
    alexa = 801
    google = 802
    ifttt = 803
    yandex = 804
    homekit = 805
    xiaoai = 806
    tmall_genie = 807
    duer = 808
    dingdong = 809
    siri = 810
    clova = 811
    wechat = 901
    alipay = 902
    aqara = 903
    hisense = 904
    huawei = 905
    widget_launch = 820
    smart_watch = 821


class RoborockDssCodes(RoborockEnum):
    @classmethod
    def _missing_(cls: type[Self], key) -> Self:
        # If the calculated value is not provided, then it should be viewed as okay.
        # As the math will sometimes result in you getting numbers that don't matter.
        return cls.okay  # type: ignore


class ClearWaterBoxStatus(RoborockDssCodes):
    """Status of the clear water box."""

    okay = 0
    out_of_water = 1
    out_of_water_2 = 38
    refill_error = 48


class DirtyWaterBoxStatus(RoborockDssCodes):
    """Status of the dirty water box."""

    okay = 0
    full_not_installed = 1
    full_not_installed_2 = 39
    drain_error = 49


class DustBagStatus(RoborockDssCodes):
    """Status of the dust bag."""

    okay = 0
    not_installed = 1
    full = 34


class CleanFluidStatus(RoborockDssCodes):
    """Status of the cleaning fluid container."""

    empty_not_installed = 1
    okay = 2


class RoborockErrorCode(RoborockEnum):
    none = 0
    lidar_blocked = 1
    bumper_stuck = 2
    wheels_suspended = 3
    cliff_sensor_error = 4
    main_brush_jammed = 5
    side_brush_jammed = 6
    wheels_jammed = 7
    robot_trapped = 8
    no_dustbin = 9
    strainer_error = 10  # Filter is wet or blocked
    compass_error = 11  # Strong magnetic field detected
    low_battery = 12
    charging_error = 13
    battery_error = 14
    wall_sensor_dirty = 15
    robot_tilted = 16
    side_brush_error = 17
    fan_error = 18
    dock = 19  # Dock not connected to power
    optical_flow_sensor_dirt = 20
    vertical_bumper_pressed = 21
    dock_locator_error = 22
    return_to_dock_fail = 23
    nogo_zone_detected = 24
    visual_sensor = 25  # Camera error
    light_touch = 26  # Wall sensor error
    vibrarise_jammed = 27
    robot_on_carpet = 28
    filter_blocked = 29
    invisible_wall_detected = 30
    cannot_cross_carpet = 31
    internal_error = 32
    collect_dust_error_3 = 34  # Clean auto-empty dock
    collect_dust_error_4 = 35  # Auto empty dock voltage error
    mopping_roller_1 = 36  # Wash roller may be jammed
    mopping_roller_error_2 = 37  # wash roller not lowered properly
    clear_water_box_hoare = 38  # Check the clean water tank
    dirty_water_box_hoare = 39  # Check the dirty water tank
    sink_strainer_hoare = 40  # Reinstall the water filter
    clear_water_box_exception = 41  # Clean water tank empty
    clear_brush_exception = 42  # Check that the water filter has been correctly installed
    clear_brush_exception_2 = 43  # Positioning button error
    filter_screen_exception = 44  # Clean the dock water filter
    mopping_roller_2 = 45  # Wash roller may be jammed
    up_water_exception = 48
    drain_water_exception = 49
    temperature_protection = 51  # Unit temperature protection
    clean_carousel_exception = 52
    clean_carousel_water_full = 53
    water_carriage_drop = 54
    check_clean_carouse = 55
    audio_error = 56


class RoborockFanPowerCode(RoborockEnum):
    """Describes the fan power of the vacuum cleaner."""

    # Fan speeds should have the first letter capitalized - as there is no way to change the name in translations as
    # far as I am aware


class RoborockFanSpeedV1(RoborockFanPowerCode):
    silent = 38
    standard = 60
    medium = 77
    turbo = 90


class RoborockFanSpeedV2(RoborockFanPowerCode):
    silent = 101
    balanced = 102
    turbo = 103
    max = 104
    gentle = 105
    auto = 106


class RoborockFanSpeedV3(RoborockFanPowerCode):
    silent = 38
    standard = 60
    medium = 75
    turbo = 100


class RoborockFanSpeedE2(RoborockFanPowerCode):
    gentle = 41
    silent = 50
    standard = 68
    medium = 79
    turbo = 100


class RoborockFanSpeedS7(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106


class RoborockFanSpeedS7MaxV(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108


class RoborockFanSpeedS6Pure(RoborockFanPowerCode):
    gentle = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106


class RoborockFanSpeedQ7Max(RoborockFanPowerCode):
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104


class RoborockFanSpeedQRevoMaster(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedQRevoCurv(RoborockFanPowerCode):
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    off = 105
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedQRevoMaxV(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedP10(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedS8MaxVUltra(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedSaros10(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockFanSpeedSaros10R(RoborockFanPowerCode):
    off = 105
    quiet = 101
    balanced = 102
    turbo = 103
    max = 104
    custom = 106
    max_plus = 108
    smart_mode = 110


class RoborockMopModeCode(RoborockEnum):
    """Describes the mop mode of the vacuum cleaner."""


class RoborockMopModeQRevoCurv(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopModeS7(RoborockMopModeCode):
    """Describes the mop mode of the vacuum cleaner."""

    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303


class RoborockMopModeS8ProUltra(RoborockMopModeCode):
    standard = 300
    deep = 301
    deep_plus = 303
    fast = 304
    custom = 302
    smart_mode = 306


class RoborockMopModeS8MaxVUltra(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    deep_plus_pearl = 305
    smart_mode = 306


class RoborockMopModeSaros10R(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopModeQRevoMaster(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopModeQRevoMaxV(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopModeSaros10(RoborockMopModeCode):
    standard = 300
    deep = 301
    custom = 302
    deep_plus = 303
    fast = 304
    smart_mode = 306


class RoborockMopIntensityCode(RoborockEnum):
    """Describes the mop intensity of the vacuum cleaner."""


class RoborockMopIntensityS7(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    mild = 201
    moderate = 202
    intense = 203
    custom = 204


class RoborockMopIntensityV2(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 207


class RoborockMopIntensityQRevoMaster(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207
    smart_mode = 209


class RoborockMopIntensityQRevoCurv(RoborockMopIntensityCode):
    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207
    smart_mode = 209


class RoborockMopIntensityQRevoMaxV(RoborockMopIntensityCode):
    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207
    smart_mode = 209


class RoborockMopIntensityP10(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207
    smart_mode = 209


class RoborockMopIntensityS8MaxVUltra(RoborockMopIntensityCode):
    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    max = 208
    smart_mode = 209
    custom_water_flow = 207


class RoborockMopIntensitySaros10(RoborockMopIntensityCode):
    off = 200
    mild = 201
    standard = 202
    intense = 203
    extreme = 208
    custom = 204
    smart_mode = 209


class RoborockMopIntensitySaros10R(RoborockMopIntensityCode):
    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    extreme = 250
    vac_followed_by_mop = 235
    smart_mode = 209


class RoborockMopIntensityS5Max(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207


class RoborockMopIntensityS6MaxV(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom = 204
    custom_water_flow = 207


class RoborockMopIntensityQ7Max(RoborockMopIntensityCode):
    """Describes the mop intensity of the vacuum cleaner."""

    off = 200
    low = 201
    medium = 202
    high = 203
    custom_water_flow = 207


class RoborockDockErrorCode(RoborockEnum):
    """Describes the error code of the dock."""

    ok = 0
    """No error condition."""

    no_dustbin_or_filter = 32
    """No dock dustbin or filter installed.

    This error message applies to auto-empty docks.
    """

    auto_empty_dock_fan_error = 33
    """Auto-Empty Dock fan error: Dock dustbin or filter out of place."""

    duct_blockage = 34
    """Auto-Empty Dock jammed: Dock dustbin, filter, or air duct jammed, check and clean it."""

    auto_empty_dock_voltage_error = 35
    """Auto-Empty Dock voltage error: Unable to empty the dustbin."""

    water_empty = 38
    """Clean water tank empty: Check tank placement or refill as required."""

    waste_water_tank_full = 39
    """Check the dirty water tank: Check tank placement or empty as required."""

    maintenance_brush_jammed = 42
    """Self-cleaning roller error: Maintenance brush jammed. Remove and clean."""

    dirty_tank_latch_open = 44
    """Water filter blocked: Clean and reinstall.

    Make sure that the dirty water tank cover is closed and the latch is secured.
    """

    no_dustbin = 46
    """Dustbin not installed (Standard error for missing dustbin).

    This error message applies to larger wash docks.
    """

    cleaning_tank_full_or_blocked = 53
    """Cleaning tank full or blocked (Water filter or sink strainer blocked/not installed)."""


class RoborockDockTypeCode(RoborockEnum):
    unknown = -9999
    no_dock = 0
    auto_empty_dock = 1
    empty_wash_fill_dock = 3
    auto_empty_dock_pure = 5
    s7_max_ultra_dock = 6
    s8_dock = 7
    p10_dock = 8
    p10_pro_dock = 9
    s8_maxv_ultra_dock = 10
    qrevo_master_dock = 14
    qrevo_s_dock = 15
    saros_r10_dock = 16
    qrevo_curv_dock = 17
    saros_10_dock = 18
    qrevo_s5v_dock = 22
    saros_20_dock = 27


class RoborockDockDustCollectionModeCode(RoborockEnum):
    """Describes the dust collection mode of the vacuum cleaner."""

    # TODO: Get the correct values for various different docks
    unknown = -9999
    smart = 0
    light = 1
    balanced = 2
    max = 4


class RoborockStateCode(RoborockEnum):
    unknown = 0
    starting = 1
    charger_disconnected = 2
    idle = 3
    remote_control_active = 4
    cleaning = 5
    returning_home = 6
    manual_mode = 7
    charging = 8
    charging_problem = 9
    paused = 10
    spot_cleaning = 11
    error = 12
    shutting_down = 13
    updating = 14
    docking = 15
    going_to_target = 16
    zoned_cleaning = 17
    segment_cleaning = 18
    emptying_the_bin = 22  # on s7+
    washing_the_mop = 23  # on a46
    washing_the_mop_2 = 25
    going_to_wash_the_mop = 26  # on a46
    in_call = 28
    mapping = 29
    egg_attack = 30
    patrol = 32
    attaching_the_mop = 33  # on g20s ultra
    detaching_the_mop = 34  # on g20s ultra
    charging_complete = 100
    device_offline = 101
    locked = 103
    air_drying_stopping = 202
    robot_status_mopping = 6301
    clean_mop_cleaning = 6302
    clean_mop_mopping = 6303
    segment_mopping = 6304
    segment_clean_mop_cleaning = 6305
    segment_clean_mop_mopping = 6306
    zoned_mopping = 6307
    zoned_clean_mop_cleaning = 6308
    zoned_clean_mop_mopping = 6309
    back_to_dock_washing_duster = 6310



"""Thin wrapper around the MQTT channel for Roborock B01 Q10 devices."""

import logging
from collections.abc import AsyncGenerator, Callable
from typing import Protocol

from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP
from roborock.devices.transport.channel import Channel
from roborock.devices.transport.mqtt_channel import MqttChannel
from roborock.exceptions import RoborockException
from roborock.protocols.b01_q10_protocol import (
    ParamsType,
    Q10Message,
    decode_message,
    encode_mqtt_payload,
)
from roborock.roborock_message import RoborockMessage

_LOGGER = logging.getLogger(__name__)


class Q10RpcChannel(Protocol):
    """Protocol for Q10 RPC channels."""

    async def send_command(
        self,
        command: B01_Q10_DP,
        params: ParamsType = None,
    ) -> None:
        """Send a command on the MQTT channel, without waiting for a response."""
        ...


class B01Q10Channel(Channel, Q10RpcChannel):
    """Unified B01 Q10 channel wrapping MQTT transport."""

    def __init__(self, mqtt_channel: MqttChannel) -> None:
        self._mqtt_channel = mqtt_channel

    @property
    def is_connected(self) -> bool:
        return self._mqtt_channel.is_connected

    @property
    def is_local_connected(self) -> bool:
        return False

    async def subscribe(self, callback: Callable[[RoborockMessage], None]) -> Callable[[], None]:
        return await self._mqtt_channel.subscribe(callback)

    async def subscribe_stream(self) -> AsyncGenerator[Q10Message, None]:
        """Stream decoded Q10 messages received via MQTT."""
        async for msg in stream_decoded_messages(self._mqtt_channel):
            yield msg

    async def send_command(
        self,
        command: B01_Q10_DP,
        params: ParamsType = None,
    ) -> None:
        await send_command(self._mqtt_channel, command, params)


async def stream_decoded_messages(
    mqtt_channel: MqttChannel,
) -> AsyncGenerator[Q10Message, None]:
    """Stream decoded Q10 messages received via MQTT.

    Each pushed ``RoborockMessage`` is decoded into a typed :data:`Q10Message`
    (a DPS status update, a map packet, or a trace packet). Messages that fail
    to decode or carry an unrecognized payload are skipped.
    """

    async for message in mqtt_channel.subscribe_stream():
        try:
            decoded = decode_message(message)
        except RoborockException as ex:
            _LOGGER.debug(
                "Failed to decode B01 Q10 message: %s: %s",
                message,
                ex,
            )
            continue
        if decoded is not None:
            yield decoded


async def send_command(
    mqtt_channel: MqttChannel,
    command: B01_Q10_DP,
    params: ParamsType,
) -> None:
    """Send a command on the MQTT channel, without waiting for a response"""
    _LOGGER.debug("Sending B01 MQTT command: cmd=%s params=%s", command, params)
    roborock_message = encode_mqtt_payload(command, params)
    _LOGGER.debug("Sending MQTT message: %s", roborock_message)
    try:
        await mqtt_channel.publish(roborock_message)
    except RoborockException as ex:
        _LOGGER.debug(
            "Error sending B01 decoded command (method=%s params=%s): %s",
            command,
            params,
            ex,
        )
        raise


def create_b01_q10_channel(mqtt_channel: MqttChannel) -> B01Q10Channel:
    """Create a B01Q10Channel wrapping MQTT transport."""
    return B01Q10Channel(mqtt_channel)



"""Trait for device network information."""

import logging

from roborock.data import NetworkInfo
from roborock.devices.cache import DeviceCache
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand

_LOGGER = logging.getLogger(__name__)


class NetworkInfoConverter(common.V1TraitDataConverter):
    """Converter for NetworkInfo objects."""

    def convert(self, response: common.V1ResponseData) -> NetworkInfo:
        """Parse the response from the device into a NetworkInfoConverter instance."""
        if not isinstance(response, dict):
            raise ValueError(f"Unexpected NetworkInfoTrait response format: {response!r}")
        return NetworkInfo.from_dict(response)


class NetworkInfoTrait(NetworkInfo, common.V1TraitMixin):
    """Trait for device network information.

    This trait will always prefer reading from the cache if available. This
    information is usually already fetched when creating the device local
    connection, so reading from the cache avoids an unnecessary RPC call.
    However, we have the fallback to reading from the device if the cache is
    not populated for some reason.
    """

    command = RoborockCommand.GET_NETWORK_INFO
    converter = NetworkInfoConverter()

    def __init__(self, device_uid: str, device_cache: DeviceCache) -> None:  # pylint: disable=super-init-not-called
        """Initialize the trait."""
        self._device_uid = device_uid
        self._device_cache = device_cache
        self.ip = ""

    async def refresh(self) -> None:
        """Refresh the network info from the cache."""

        device_cache_data = await self._device_cache.get()
        if device_cache_data.network_info:
            _LOGGER.debug("Using cached network info for device %s", self._device_uid)
            common.merge_trait_values(self, device_cache_data.network_info)
            return

        # Load from device if not in cache
        _LOGGER.debug("No cached network info for device %s, fetching from device", self._device_uid)
        await super().refresh()

        # Update the cache with the new network info
        device_cache_data = await self._device_cache.get()
        device_cache_data.network_info = self
        await self._device_cache.set(device_cache_data)



from unittest.mock import patch

import pytest
from vacuum_map_parser_base.map_data import MapData

from roborock.devices.traits.b01.q7 import Q7PropertiesApi
from roborock.exceptions import RoborockException
from roborock.map.b01_map_parser import ParsedMapData
from roborock.roborock_typing import RoborockB01Q7Methods

from .conftest import FakeQ7Channel


async def test_q7_map_content_refresh_populates_cached_values(
    q7_api: Q7PropertiesApi,
    fake_channel: FakeQ7Channel,
):
    fake_channel.response_queue.extend(
        [
            {"map_list": [{"id": 1772093512, "cur": True}]},
            b"inflated-payload",
        ]
    )

    # Ensure we have map metadata first
    await q7_api.map.refresh()

    dummy_map_data = MapData()
    parsed_map_data = ParsedMapData(
        image_content=b"pngbytes",
        map_data=dummy_map_data,
    )
    with patch(
        "roborock.devices.traits.b01.q7.map_content.B01MapParser.parse",
        return_value=parsed_map_data,
    ) as parse:
        await q7_api.map_content.refresh()

    assert q7_api.map_content.image_content == b"pngbytes"
    assert q7_api.map_content.map_data is dummy_map_data
    assert q7_api.map_content.raw_api_response == b"inflated-payload"

    parse.assert_called_once_with(b"inflated-payload")

    assert len(fake_channel.published_commands) == 2
    cmd, params = fake_channel.published_commands[0]
    assert cmd == RoborockB01Q7Methods.GET_MAP_LIST

    map_cmd, map_params = fake_channel.published_commands[1]
    assert map_cmd == RoborockB01Q7Methods.UPLOAD_BY_MAPID
    assert map_params == {"map_id": 1772093512}


async def test_q7_map_content_refresh_falls_back_to_first_map(
    q7_api: Q7PropertiesApi,
    fake_channel: FakeQ7Channel,
):
    """If no current map marker exists, first map in list is used."""
    fake_channel.response_queue.extend(
        [
            {"map_list": [{"id": 111}, {"id": 222, "cur": False}]},
            b"inflated-payload",
        ]
    )

    # Load current map
    await q7_api.map.refresh()

    dummy_map_data = MapData()
    with patch(
        "roborock.devices.traits.b01.q7.map_content.B01MapParser.parse",
        return_value=type("X", (), {"image_content": b"pngbytes", "map_data": dummy_map_data})(),
    ):
        await q7_api.map_content.refresh()

    assert len(fake_channel.published_commands) == 2
    map_cmd, map_params = fake_channel.published_commands[1]
    assert map_params == {"map_id": 111}


async def test_q7_map_content_refresh_errors_without_map_list(
    q7_api: Q7PropertiesApi,
    fake_channel: FakeQ7Channel,
):
    """Refresh should fail clearly when map list is unusable."""
    fake_channel.response_queue.extend([{"map_list": []}])

    with pytest.raises(RoborockException, match="Unable to determine current map ID"):
        await q7_api.map_content.refresh()
